# Native Modules
import gzip
from subprocess import Popen, PIPE
import os
# Installed Modules
from Bio.Seq import Seq
# Project Modules


class Transcript:
    tx_lib = {} #tid : []
    coord2tid = {}
    labels = ['chrom', 'txStart', 'txEnd', 'strand', 'tid', 'eid', 'name',
              'cdsStart', 'cdsEnd', 'exonStarts', 'exonEnds', 'exonFrames']

    def __init__(self,entry):
        '''
        :param entry: NCBI Transcript Entry
        '''

        self.entry = entry
        for k in ['txStart', 'txEnd','cdsStart', 'cdsEnd']:
            self.entry[k] = int(self.entry[k])

        self.overlapping_transcripts = []

        ##output
        #mapping transcript features (relative to transcript start)
        self.exons= self.get_exons()
        self.utrs = self.get_utrs()
        self.tx_len = self.entry['txEnd'] - self.entry['txStart']  # 1608
        self.flanking = [(-50, 0), (self.tx_len, self.tx_len + 50)]

        self.feature = None
        self.rf = None
        self.cdsseq = None
        self.txseq = None

    @classmethod
    def transcript(cls, snvcoord):
        if snvcoord in cls.coord2tid.keys():
            tid = cls.coord2tid[snvcoord]
            start, end = snvcoord.split(':')[1].split('-')
            pos = int((int(start) + int(end)) /2)
            obj = cls.tx_lib[tid]
            obj.find_feature(pos)
            return obj
        else:
            return 'intergenic'

    @classmethod
    def load_transcripts(cls, annote_path,snvcoords):
        # print('DANIEL WE MIGHT NEED TO MAKE A TEMP FOLDER')
        # print('SEE load_transcripts in annotate.py')
        temp_bedfname = "temp_in.bed"
        bed_entries = None

        #reset dict
        cls.coord2tid = {}
        cls.tx_lib = {}

        with open(temp_bedfname, 'w') as tempbed:
            for coord in snvcoords:
                coord_field = coord.split(':')
                chrom = coord_field[0]
                start = coord_field[1].split('-')[0]
                end = coord_field[1].split('-')[1]
                line = "\t".join([chrom, start, end])
                tempbed.writelines(line + "\n")

        cmd = 'bedtools window -w 50 -a ' + temp_bedfname + ' -b ' + annote_path
        bed_popen = Popen(cmd, shell=True, stdout=PIPE, text=True)
        bedtools_out = bed_popen.communicate()[0]
        ret = bed_popen.wait()
        if ret != 0:
            print("Bedtools process was interrupted!")
        if bedtools_out != '':
            bed_entries = bedtools_out.split('\n')[:-1]

            for bed_entry in bed_entries:
                tokens = bed_entry.split('\t')
                tokens = tokens[:-1] + tokens[-1].split('|')
                snvcoord = f'{tokens[0]}:{tokens[1]}-{tokens[2]}'
                entry = dict(zip(cls.labels, tokens[-12:]))

                if entry['chrom'] != '.':
                    if snvcoord not in cls.coord2tid.keys():
                        cls.coord2tid[snvcoord] = entry['tid']
                        cls.tx_lib[entry['tid']] = cls(entry)
                    # adjusting for overlaps to retrieve the most up to date transcript
                    else:
                        new_tid = entry['tid']
                        oldtid = cls.coord2tid[snvcoord]

                        if new_tid.startswith('NR') and oldtid.startswith('NM'):
                            pass

                        elif new_tid[0:2] == oldtid[0:2]:
                            if entry['eid'] != '-' or (float(oldtid.split('_')[-1]) < float(new_tid.split('_')[-1])):
                                cls.coord2tid[snvcoord] = new_tid
                                obj = cls(entry)
                                obj.overlapping_transcripts.append(oldtid)
                                cls.tx_lib[entry['tid']] = obj
                        else:
                            cls.coord2tid[snvcoord] = new_tid
                            obj = cls(entry)
                            obj.overlapping_transcripts.append(oldtid)
                            cls.tx_lib[entry['tid']] = obj

                else:
                    cls.coord2tid[snvcoord] = 'intergenic'
        os.remove(temp_bedfname)
    def __dict__(self):
        return self.entry
    def get_utrs(self):
        exon_starts = self.entry['exonStarts'][:-1].split(',')
        exon_ends = self.entry['exonEnds'][:-1].split(',')
        if self.exons:
            ogexons = [(int(exon_starts[i]) - self.entry['txStart'], int(exon_ends[i]) - self.entry['txStart']) for i in range(len(exon_ends))]
            utrs= [(ogexons[0][0],self.exons[0][0]-1),(self.exons[-1][1],ogexons[-1][1]-1)]
        else:
            utrs = None
        return utrs

    def get_exons(self):
        '''
        uses entry info to find relative exons positions
        '''
        exon_starts = self.entry['exonStarts'][:-1].split(',')
        exon_ends = self.entry['exonEnds'][:-1].split(',')
        exon_frames = self.entry['exonFrames'].replace("\n", "")[:-1].split(',')
        tx_start = self.entry['txStart']

        exons = [(int(exon_starts[i]) - tx_start, int(exon_ends[i]) - tx_start) for i in range(len(exon_ends))]
        for i in range(len(exon_frames)):
            if exon_frames[i] == '-1':  # -1 means entire exon is UTR
                exons = exons[1:]
                exon_starts = exon_starts[1:]
            else:
                break
        for i in range(1, len(exon_frames)):
            if exon_frames[-i] == '-1':
                exons = exons[0:len(exons) - 1]
                exon_ends = exon_ends[0:-1]
            else:
                break

        # Determine the stop and start of UTR
        if len(exons) > 0:
            exons[0] = (int(self.entry['cdsStart']) - int(exon_starts[0]) + exons[0][0], exons[0][1])
            exons[-1] = (exons[-1][0], exons[-1][1] - (int(exon_ends[-1]) - int(self.entry['cdsEnd'])))
        else:
            exons = None
        return exons


    def get_tx_seq(self,fasta_seq):
        '''
        Using a Refseq Transcript_ID, Ensembl Transcript_ID or coordinates find transcript annotations and transcript sequence
        from either a genome fasta path or given genome sequence
        '''
        self.tx_seq = fasta_seq.seq[self.entry['txStart']:self.entry['txEnd']]
        return self.tx_seq

    def get_cdsseq(self):
        '''
        uses entry info to adjust exons positions relative to transcription start
        removes utrs from exons
        translates into cds
        '''
        # Determine the stop and start of UTR
        if self.tx.seq:
            cds = Seq(''.join([str(self.tx_seq)[a:b] for a, b in self.exons]))
            if self.entry['strand'] == '-':
                cds = cds.reverse_complement()
        else:
            cds = None
        return cds


    def tx_info(self):
        return self.entry['eid'], self.entry['tid'], self.entry['name'], self.entry['strand'], self.entry['txStart']


    def find_reading_frame(self, dist_from_cds_start):
        '''
        Finds reading frame of SNV in extracted sequence
        '''
        rf = 1 if dist_from_cds_start % 3 == 2 else 2 if dist_from_cds_start % 3 == 0 else 0
        return rf

    def find_feature(self, pos):
        feature, rf = None, None
        t_snvpos = int(pos) - self.entry['txStart']
        #cdstart, cdsend = self.cds_start, self.cds_end

        if self.entry['tid'].startswith('NR'):
            feature = 'non-coding RNA'

        elif self.exons == None or t_snvpos < -50 or t_snvpos > self.tx_len + 50:
            # not in transcript - shouldn't happen or else no entry would be found
            feature = 'non-coding'

        elif t_snvpos in range(self.flanking[0][0],self.flanking[0][1]+1) or t_snvpos in range(self.flanking[1][0],self.flanking[1][1]+1):

            if self.entry['strand']=='+' and t_snvpos in range(self.flanking[0][1] -25,self.flanking[0][1] -36):
                feature = 'flanking - upstream - promoter'
            elif self.entry['strand']=='-' and t_snvpos in range(self.flanking[1][1] +25,self.flanking[1][1] +36):
                feature = 'flanking - upstream - promoter'
            elif t_snvpos in range(self.flanking[0][0],self.flanking[0][1]+1):
                feature = 'flanking-upstream' if self.entry['strand'] == '+' else 'flanking-downstream'
            else:
                feature = 'flanking-downstream' if self.entry['strand'] == '+' else 'flanking-downstream'


        elif t_snvpos in range(self.utrs[0][0],self.utrs[0][1]+1) or t_snvpos in range(self.utrs[1][0],self.utrs[1][1]+1):

            if t_snvpos in range(self.utrs[0][0],self.utrs[0][1]+1):
                feature = '5utr' if self.entry['strand'] == '+' else '3utr'
            else:
                feature = '3utr' if self.entry['strand'] == '+' else '5utr'


        else:# find if exon or intron
            feature = 'intron'
            exon_n = 0
            for x in self.exons:
                # if in exon find reading frame
                if t_snvpos in range(x[0], x[1] + 1):
                    feature = 'exon'
                    dist = sum([e[1] - e[0] for e in self.exons[0:exon_n]])
                    dist_from_cds_start = dist + (t_snvpos - x[0])
                    len_cds = sum([e[1] - e[0] for e in self.exons])

                    if self.entry['strand'] == '-':
                        dist_from_cds_start = (len_cds - dist_from_cds_start) + 1

                    if dist_from_cds_start < 3:
                        feature = 'start_codon'

                    if dist_from_cds_start > len_cds - 3:
                        feature = 'stop_codon'

                    rf = self.find_reading_frame(dist_from_cds_start)

                    break
                exon_n += 1

        self.feature, self.rf = feature, rf


'''test
annote_path = '/groups/clinical/projects/editability/tables/processed_tables/ncbiRefSeq.bed.gz'
snvcoords = ['chr11:5225460-5225460','chr11:5226676-5226676',
             'chr3:136250375-136250375','chr16:57744390-57744390',
             'chr16:136330169-136330169','chr18:58671560-58671560'
             ]
Transcript.load_transcripts(annote_path,snvcoords)
for snv in snvcoords:
    pos_in_transcript = Transcript.transcript(snv)
    if pos_in_transcript != 'intergenic':
        print(pos_in_transcript.tx_info())
        print(pos_in_transcript.feature)
    else:
        print('intergenic', snv)
#('ENST00000335295.4', 'NM_000518.5', 'HBB', '-', 5225463)
#flanking-downstream
#('ENST00000335295.4', 'NM_000518.5', 'HBB', '-', 5225463)
#exon
#('-', 'NM_001178014.2', 'PCCB', '+', 136250339)
#start_codon
#('ENST00000379661.8', 'NM_005886.3', 'KATNB1', '+', 57735769)
#intron
#('ENST00000649217.2', 'NM_006785.4', 'MALT1', '+', 58671464)
#5utr


'''

#    def get_refseq_entry(term, field):
#        '''
#        Using ncbiRefSeq.bed.gz to find cds features by either interval, gene name or transcript ID
#        example input:
#        term, field = 'NM_000532.5', 'tid'
#        term, field = 'PCCB','name'
#        term,field =  'chr3:136250339-136330169','interval'
#        annote_path = '/groups/clinical/projects/editability/tables/processed_tables/ncbiRefSeq.bed.gz'
#        '''
#
#        labels = ['chrom', 'txStart', 'txEnd','strand','tid','eid,','name',
#                  'cdsStart', 'cdsEnd','exonStart','exonEnd', 'exonFrames']
#
#        if field != 'interval':
#            not_found = True
#            for line in gzip.open(annote_path, 'rt'):
#                tokens = line.split('\t')
#                entry = dict(zip(labels, tokens))
#                if term in entry[field]:
#                    not_found = False
#                    break

#            if not_found:
#                entry = None
#                print(f"{term} not found in refseq data")

 #               if '.' in term:
 #                   new_term = term.split('.')[0]
 #                   print(f'searching for {new_term} instead')
 #                   entry = get_refseq_entry(new_term, field,annote_path)

 #       else:  # only used for intervals search
 #           not_found = True
 #           ch = term.split(":")[0]
 #           start, end = term.split(":")[1].split('-')
 #           pos = int((int(start) + int(end)) / 2)

 #           for line in gzip.open(annote_path, 'rt'):
 #               tokens = line.split('\t')
 #               entry = dict(zip(labels, tokens))
 #               if ch == entry['chrom']:
 #                   if pos in range(int(entry['txStart']), int(entry['txEnd'])):
 #                       not_found = False
 #                       break
 #           if not_found:
 #               entry = None
 #               print(f"{term} not found in refseq data")
 #       return entry




# Native modules
from datetime import datetime, date
import subprocess
import gzip
import regex as re
# Installed Modules
import pandas as pd
import yaml

# print(f"Loading configuration file {config_path}")
# 	with open(config_path, "r") as f:
# 		config = yaml.load(f, Loader=yaml.FullLoader)

'''
clinvar_ftp = "https://ftp.ncbi.nlm.nih.gov/pub/clinvar/tab_delimited/variant_summary.txt.gz"
clinvar_vcf = ' https://ftp.ncbi.nlm.nih.gov/pub/clinvar/vcf_GRCh38/clinvar.vcf.gz'
clinvar_index = ' https://ftp.ncbi.nlm.nih.gov/pub/clinvar/vcf_GRCh38/clinvar.vcf.gz.tbi'
refseq = "https://hgdownload.soe.ucsc.edu/goldenPath/hg38/database/ncbiRefSeq.txt.gz"
'''
datadir = "/groups/clinical/projects/editability/tables/"

# database paths
raw_tables = f"{datadir}raw_tables/"
clinvar_summary = f"{raw_tables}clinvar/variant_summary.txt.gz"  # raw clinvar table
mane_path = f'{raw_tables}clinvar/MANE.GRch38.v1.1.summary.txt.gz'
HPApath = f"{raw_tables}HPA/proteinatlas.tsv"


processed_tables = f"{datadir}processed_tables/"
simple_tables = f"{processed_tables}guide_acquisition_tables/"
HGVSlookup_path = f"{processed_tables}HGVSlookup.csv" #Chrom to refID key table
lastUpdate_file = f"{processed_tables}clinvar_lastUpdate.txt"
mane_cleaned = f'{processed_tables}MANE.GRch38.summary_cleaned.txt.gz'



#variables
chroms = ['1', '2','3', '4', '5', '6', '7', '8', '9', '10', '11','12',
                 '13', '14', '15', '16', '17', '18', '19', '20', '21', '22','MT', 'Y', 'X']


def process_refseq(raw_tables,processed_tables):
    '''
    ftp : https://hgdownload.soe.ucsc.edu/goldenPath/hg38/database/ncbiRefSeq.txt.gz

    Makes
    A) a bed file from refseqs = 282,614 lines
    B) a bed file of the most recent genes = 66688 lines
    '''
    #in
    mane = pd.read_csv(f'{raw_tables}clinvar/MANE.GRch38.v1.1.summary.txt.gz', sep="\t")

    #out
    ref_out = gzip.open(f'{processed_tables}ncbiRefSeq.bed.gz', 'wt')

    labels = ['bin', 'id', 'chrom', 'strand', 'txStart', 'txEnd',
              'cdsStart', 'cdsEnd', 'exonCount', 'exonStarts', 'exonEnds',
              'score', 'name', 'cdsStartStat', 'cdsEndStat','exonFrames']

    mane_ensid = list(mane['Ensembl_nuc'])
    mane_tid = list(mane['RefSeq_nuc'])

    latest_refs = {}
    cnt = 0
    for line in gzip.open(f'{raw_tables}Refseq/ncbiRefSeq.txt.gz', 'rt'):
        tokens = line.split('\t')
        tid, chrom, strand,tstart,tend = tokens[1:6]
        cds_start, cds_end = tokens[6], tokens[7]
        exons_start, exon_end = tokens[9], tokens[10]
        gname, frames = tokens[12], tokens[-1].split('\n')[0]

        if tokens[2].replace('chr', "") in chroms:
            tid = tokens[1]
            cnt+=1
            if tid.startswith('X') == False:
                try:
                    i = mane_tid.index(tid)
                    eid = mane_ensid[i]
                except:

                    eid = '-'
                line_out = [chrom,tstart,tend,'|'.join([strand,tid,eid,gname,cds_start,cds_end, exons_start,exon_end,frames])]
                ref_out.write('\t'.join(line_out) + '\n')
                if cnt < 20:
                    print('\t'.join(line_out))

    ref_out.close()


def process_MANE(raw_tables,processed_tables):
    '''
    Mane has joins ENS and REFSEQ IDs
    '''
    mane = pd.read_csv(f'{raw_tables}clinvar/MANE.GRch38.v1.1.summary.txt.gz', sep="\t")

    mane['#NCBI_GeneID'] = mane['#NCBI_GeneID'].str.replace('GeneID:',"")
    mane = mane[mane['MANE_status'] == 'MANE Select']
    mane = mane.rename(columns = {'#NCBI_GeneID': 'GeneID','symbol':'GeneSymbol','RefSeq_nuc':'TranscriptID',
                         'RefSeq_prot':'ProteinID','chr_start':'Start',
                         'chr_end':'End','chr_strand':'Strand','GRCh38_chr':'ChrID','Ensembl_nuc':'Ensembl_TranscriptID'})

    mane = mane.drop(columns = 'MANE_status')
    mane = mane.loc[mane.ChrID.str.startswith('NC_')]
    mane.to_csv(f'{processed_tables}MANE.GRch38.summary_cleaned.txt.gz',index = False, compression='gzip')


def process_clinvarVCF(processed_tables):
    '''
    clinvarvcf has molecular consequences
    '''
    mc = pd.read_csv('/groups/clinical/projects/editability/tables/raw_tables/clinvar/clinvar.vcf'
                     , comment='#', sep='\t', names=['Chr', 'PositionVCF', '?', 'REF', 'ALT', 'x', 'x1','attributes'])
    mc['chrHGVS_ID'] = mc['attributes'].str.extract(r'CLNHGVS=([^;]*)', expand=False)
    mc['MC'] = mc['attributes'].str.extract(r'MC=([^;]*)', expand=False)
    mc['AlleleID'] = mc['attributes'].str.extract(r'ALLELEID=([^;]*)', expand=False)
    con = []
    mc.loc[mc['MC'].isna(),'MC'] = '-'
    for x in mc['MC']:
        if '|' in x:
            x = ','.join([y for y in x.split('|') if 'SO' not in y])
        con.append(x)
    mc['MC'] = con

    mc = mc.drop(columns=['x1', 'x', 'attributes'])
    mc.to_csv(f'{processed_tables}clinvarvcf2txt.txt', index=False)


def add_molecular_consequences(chroms,processed_tables):
    '''
    add molecular consequences
    '''
    for ch in chroms[10:]:
        df = pd.read_csv(f"{processed_tables}variant_tables/{ch}_variant.txt")
        mc = pd.read_csv(f'{processed_tables}clinvarvcf2txt.txt')
        mc['Chr'] = mc['Chr'].astype('str')
        mc = mc[mc['Chr'] == ch]
        mc = mc[['MC', 'AlleleID']]
        mc['AlleleID'] = mc['AlleleID'].astype('int64')
        joined = df.join(mc.set_index('AlleleID'), on='AlleleID')
        joined.to_csv(f'{processed_tables}variant_tables/{ch}_variant.txt', index=False)


def extract_tid_from_hgvs(vdf):
    rprefix = r"((N(M|G|C|R)_[\d.]*)|(m))"
    rsuffix = r"(:(c|m|g|n)\.\S*)"
    hgs = list(vdf['HGVS_ID'])
    simple_ids = []
    tids = []
    for h in hgs:
        if re.search(rsuffix,h) and re.search(rprefix,h):
            tid = re.search(rprefix,h).groups()[0]
            suf = tid + re.search(rsuffix,h).groups()[0]
            tids.append(tid)
            simple_ids.append(suf)
        elif h.startswith('m'):
            tid = 'm'
            suf = h
            tids.append(tid)
            simple_ids.append(suf)
        else:
            print(h)
            tids.append('-')
            simple_ids.append('-')

    vdf.insert(3, 'HGVS_Simple', simple_ids)
    vdf.insert(4,'TranscriptID',tids)
    vdf = vdf.loc[vdf['HGVS_Simple'] !='-']
    vdf = vdf.reset_index(drop=True)
    return vdf


def add_MANE(processed_tables,chroms):
    mane = pd.read_csv(f'{processed_tables}MANE.GRch38.summary_cleaned.txt.gz')
    mane = mane[['Start', 'End', 'Strand', 'TranscriptID', 'ProteinID', 'Ensembl_TranscriptID', 'Ensembl_Gene']]

    for ch in chroms:
        vdf = pd.read_csv(f"{processed_tables}variant_tables/{ch}_variant.txt")
        #mane['Ensembl_Gene'] = [x.split('.')[0] if type(x) == str else x for x in mane['Ensembl_Gene']]
        joined = vdf.join(mane.set_index('TranscriptID'), on = 'TranscriptID')

        joined.to_csv(f"{processed_tables}variant_tables/{ch}_variant.txt", index=False)
    #colreorder = [2,4,16,29,30,31,25,26,0,1,3,5,6,7,8,9,10,11,12,13,14,15,17,18,19,20,21,22,23,24,27,28,32,33,34]
    #joined = joined.iloc[:,colreorder]

def extractOMIM(vdf):
    '''
    Find OMIM ID clinvar table and make seperate column
    '''
    all_ids = [",".join([x, y]) if len(",".join([x, y])) > 0 else "NA" for x, y in
               zip(vdf['PhenoIDS'], vdf['OtherIDs'])]
    omim = []
    new_all_ids = []
    for x in all_ids:
        x = x.replace("MONDO:MONDO:", "MONDO:")
        found = list(set(re.findall("OMIM:([PS\.0-9]+)", x)))
        if len(found) == 0:
            omim.append("-")
        elif len(found) == 1:
            omim.append(found[0])
            x = x.replace(f"OMIM:{found[0]}", "")
        else:
            omim.append("|".join([z for z in found]))
            for z in found:
                x = x.replace(f"OMIM:{z}", "")
        new_all_ids.append(x)
    vdf["OMIM"] = omim
    vdf["IDs"] = new_all_ids
    vdf = vdf.drop(columns=['PhenoIDS', 'OtherIDs'])
    return vdf


def clean_clinvar(clinvar_summary,chroms):
    '''
    splits clinvar into files by chromosomes,
    removes unneeded columns
    Keeps only data from hg38 assembly
    '''
    ## Dropped cols
    # "LastEvaluated", "RS(dbSNP)", "Origin", 'Chromosome',
    # , "Cytogenetic", "ReviewStatus",
    # "NumberSubmitters", "Gudelines", "TestedInGTR", "SubmitterCategories"
    to_drop = ['RefAllele', 'AltAllele','Assembly', 'Start', 'Stop']
    in_file = gzip.open(clinvar_summary, "rt")
    contents = in_file.readlines()

    allcols = ['AlleleID', 'Type', 'HGVS_ID', 'GeneID', 'GeneSymbol', 'HGNC_ID',
               'ClinicalSign', "ClinSigSimple", "LastEval", "RS#(dbSNP)", "nsv/esv (dbVar)",
               "RCVaccession", "PhenoIDS","PhenoList", "Origin", "OriginSimple", "Assembly",
               "ChrAccession", "Chr", "Start","Stop","RefAllele", "AltAllele",
               "Cytogenetic", "ReviewStatus", "NumberSubmitters", "Guidelines", "TestedInGTR", "OtherIDs",
               "SubmitterCategories", "VariationID", "PositionVCF", "RefAlleleVCF", "AltAlleleVCF"]

    cols = [allcols.index(c) for c in allcols if c not in to_drop]
    for ch in chroms:
        print(ch)
        out_fname = f"{processed_tables}/variant_tables/{ch}_variant.txt"
        lines = []
        for line in contents:
            line = line.split("\t")
            if line[18] == str(ch):
                if line[16] != "GRCh37": # Remove hg19 data
                    line[-1] = line[-1].replace("\n", "")
                    if 'single nucleotide variant' in line:
                        line = [line[i] for i in cols]
                        lines.append(line)

        vdf = pd.DataFrame(lines, columns = [allcols[i] for i in cols])
        vdf['HGNC_ID'] = vdf['HGNC_ID'].apply(lambda x: x.replace("HGNC:",""))
        vdf['PositionVCF'] = vdf['PositionVCF'].astype('int')
        vdf = vdf[vdf['PositionVCF']>1]
        vdf = vdf.sort_values('GeneSymbol')

        #Extract GeneID and GeneSymbol from HGVSID, In places that are missing
        blankgenes = vdf.loc[vdf['GeneID'] == '-1']
        hgvs = blankgenes['HGVS_ID']
        genenames = hgvs.str.extract(r"\((\w*\d*)\):").sort_values(0)
        temp = vdf.loc[~vdf['HGVS_ID'].isin(blankgenes)].drop_duplicates(subset='GeneSymbol')
        genenames = genenames.join(temp[['GeneID','GeneSymbol']].set_index('GeneSymbol'),on = 0)
        vdf.loc[vdf['HGVS_ID'].isin(hgvs),'GeneSymbol'] = genenames[0]
        vdf.loc[vdf['HGVS_ID'].isin(hgvs),'GeneID'] = genenames['GeneID']
        vdf.loc[vdf.GeneID.isna(),'GeneID'] = -1
        vdf['GeneID'] = vdf['GeneID'].astype('int64')
        vdf = vdf.reset_index(drop=True)

        # Find and extract OMIM ID
        vdf = extract_tid_from_hgvs(vdf)
        vdf = extractOMIM(vdf)
        vdf.to_csv(out_fname,index=False)


def make_var_table(clinvar_summary,chroms):
    clean_clinvar(clinvar_summary, chroms)
    add_MANE(processed_tables, chroms)
    add_molecular_consequences(chroms,processed_tables)


def make_gene_tables(processed_tables,HPApath):
    '''
    attached HPA gene expression info to clinvar info by ensembl id
    '''

    hpa = pd.read_csv(HPApath, delimiter='\t')

    cols = ['Ensembl', 'Chromosome', 'Gene description', 'Protein class', 'Biological process',
            'Molecular function', 'Uniprot',
            'Disease involvement', 'RNA tissue specificity', 'RNA tissue specific nTPM',
            'RNA tissue distribution', 'RNA tissue cell type enrichment',
            'RNA single cell type specific nTPM', 'RNA tissue specific nTPM']
    hpa = hpa[cols]
    hpa = hpa.rename(columns = {'Ensembl':'Ensembl_Gene'})
    mane = pd.read_csv(f'{processed_tables}MANE.GRch38.summary_cleaned.txt.gz')
    mane = mane[['TranscriptID','Start', 'End', 'Strand',  'ProteinID', 'Ensembl_TranscriptID', 'Ensembl_Gene']]
    mane['Ensembl_Gene'] = [str(x).split('.')[0] for x in mane['Ensembl_Gene']]
    joined_df = hpa.join(mane.set_index('Ensembl_Gene'), on='Ensembl_Gene')
    joined_df.to_csv(f'{processed_tables}/gene_tables/gene_tables.csv.gz', index=False,compression='gzip')

def make_HGVStable(processed_tables,chroms):
    '''
    Create CSV file of unduplicated HGVS prefix(coding ref name) and Chromosome
    '''
    names, chrs = [], []
    for ch in chroms:
        clin = pd.read_csv(f"{processed_tables}variant_tables/{ch}_variant.txt")
        names += list(set(clin['TranscriptID']))
        chrs += [ch for i in range(len(set(clin['TranscriptID'])))]

    df = pd.DataFrame({"TranscriptID": names, "Chr": chrs}).drop_duplicates()
    out_file = f"{processed_tables}HGVSlookup.csv"
    df.to_csv(out_file, index=None)


def updateTables(clinvar_ftp,clinvar_summary,refseq):
    # TODO: add user inquiry to whether they want to init update
    print("updating to latest version of clinvar")

    cmd = f"wget {clinvar_ftp} -O {clinvar_summary}"
    p = subprocess.run(cmd, shell=True,
                       capture_output=True)
    print(p)

    cmd = f"wget {refseq} -O {raw_tables}Refseq/ncbiRefSeq.txt.gz"
    p = subprocess.run(cmd, shell=True,
                       capture_output=True)
    print(p)

    cmd = f"wget https://ftp.ncbi.nlm.nih.gov/refseq/MANE/MANE_human/current/MANE.GRCh38.v1.1.summary.txt.gz -O " \
          f"{raw_tables}clinvar/MANE.GRch38.v1.1.summary.txt.gz"
    p = subprocess.run(cmd, shell=True,
                       capture_output=True)
    print(p)

    # cmd = f"wget {clinvar_vcf} -O {raw_tables}clinvar/clinvar.vcf.gz"
    # p = subprocess.run(cmd, shell=True,
    #                    capture_output=True)
    # print(p)
    # cmd = f"wget {clinvar_index} -O {raw_tables}clinvar/clinvar.vcf.gz.tbi"
    # p = subprocess.run(cmd, shell=True,
    #                    capture_output=True)
    #
    # cmd = f"bcftools view -f type!=snp {clinvar_vcf} -o {raw_tables}clinvar/clinvar.vcf -O v"
    # p = subprocess.run(cmd, shell=True,
    #                    capture_output=True)

    print("Cleaning and Splitting Clinvar.....")
    clean_clinvar(clinvar_summary, chroms)
    print("Adding Ensembl Identifiers....")
    add_MANE(processed_tables, chroms)
    print("Adding Molecular Consequences....")
    add_molecular_consequences(chroms, processed_tables)
    print("Appending HPA data.....")
    make_gene_tables(processed_tables, HPApath)
    print("Writing new HGVS Lookup table.....")
    make_HGVStable(processed_tables, chroms)


    with open(lastUpdate_file, "w") as f:
        today = date.today()
        f.write(str(today))
    f.close()


def check_updates(lastUpdate_file):
    '''
    Determines if an update is needed based on checking the date of txt file
    '''

    f = open(lastUpdate_file, "r").readlines()
    lastdate = datetime.strptime(f[0], '%Y-%m-%d').date()
    if (date.today() - lastdate).days > 31:
        updateTables()
    else:
        print('You are using the latest clinvar data')
        print(f"Last updated {lastdate}")



# Native Modules
from math import exp
import pickle
import re
# Installed Modules
import pandas as pd
import numpy as np
from tensorflow.keras.models import load_model
# Project Modules
from featurization import featurize_data


def load_model_params(score_name: str, models_dir: str):
    try:
        # dirname = os.path.dirname(os.path.realpath(__file__)).replace('py/','')
        dirname = '/home/thudson/projects/editability'
        if score_name == 'cfd':
            mm_scores = pickle.load(open(models_dir + '/mismatch_score.pkl', 'rb'))
            pam_scores = pickle.load(open(models_dir + '/PAM_scores.pkl', 'rb'))
            return mm_scores, pam_scores
        if score_name == 'azimuth':
            model = pickle.load(open(models_dir + '/python3_V3_model_no.pos.pickle', 'rb'))
            return model
        if score_name == 'doench14':
            doench14params = pickle.load(open(models_dir + '/doench14params.pkl', 'rb'))
            return doench14params
        if score_name == 'deepcpf1':
            model1 = load_model(models_dir + "/DeepCpf1.h5")
            model2 = load_model(models_dir + "/Seq_deepCpf1.h5")
            return model1, model2
    except FileNotFoundError:
        raise Exception(f"File containing {score_name} could not be opened")


def revcom(s):
    basecomp = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A','U':'A'}
    letters = list(s[::-1])
    letters = [basecomp[base] for base in letters]
    return ''.join(letters)


# === On-target Efficiency Scoring ===
def doench2014(cas9_sites, models_dir):
        """
        Doench 2014 'on-target score'
        Input is a 30mer: 4bp 5', 20bp guide, 3bp PAM, 3bp 5'
        Doench et al, Nat Biotech 2014, PMID 25184501, http://www.broadinstitute.org/rnai/public/analysis-tools/sgrna-design
        Code from crispor - https://github.com/maximilianh/crisporWebsite
        """
        doench2014params = load_model_params('doench14', models_dir)
        scores = []
        global gcWeight
        intercept = 0.59763615
        gcHigh = -0.1665878
        gcLow = -0.2026259

        for seq in cas9_sites:
            assert (len(seq) == 30)
            score = intercept
            guideSeq = seq[4:24]
            gcCount = guideSeq.count("G") + guideSeq.count("C")

            if gcCount <= 10:
                gcWeight = gcLow
            if gcCount > 10:
                gcWeight = gcHigh

            score += abs(10 - gcCount) * gcWeight

            for pos, modelSeq, weight in doench2014params:
                subSeq = seq[pos:pos + len(modelSeq)]
                if subSeq == modelSeq:
                    score += weight
            scores.append(int(100 * (1.0 / (1.0 + exp(-score)))))

        return scores


def azimuth(cas9_sites, models_dir):
    '''
    Doench/Fusi 2016 Rule -2 on-target / efficiency score now packaged as 'Azimuth'
    This script is copied and modified from https://github.com/MicrosoftResearch/Azimuth
    predicts whether a guide exhibits strong or weak cleavage
    Score range 0-100. A score higher than 55% is recommended
    '''
    #Doench/Fusi 2016 Rule -2 'on-target score'
    # This script is copied and modified to suit from https://github.com/MicrosoftResearch/Azimuth
    model = load_model_params('azimuth', models_dir)
    model, learn_options = model

    res = []
    for seq in cas9_sites:
        if "N" in seq:
            res.append(-1)  # can't do Ns
            continue

        pam = seq[25:27]
        if pam != "GG":
            # res.append(-1)
            # continue
            seq = list(seq)
            seq[25] = "G"
            seq[26] = "G"
            seq = "".join(seq)
        res.append(seq)
    seqs = np.array(res)

    learn_options["V"] = 2

    Xdf = pd.DataFrame(columns=['30mer', 'Strand'],
                       data=zip(seqs, np.repeat('NA', seqs.shape[0])))
    gene_position = pd.DataFrame(columns=['Percent Peptide', 'Amino Acid Cut position'],
                                 data=zip(np.ones(seqs.shape[0]) * -1,
                                          np.ones(seqs.shape[0]) * -1))

    feature_sets = featurize_data(data=Xdf, learn_options=learn_options, Y=pd.DataFrame())
    keys = list(feature_sets.keys())
    N = feature_sets[list(keys)[0]].shape[0]
    inputs = np.zeros((N, 0))
    feature_names = []
    dim = {}
    dimsum = 0
    for set in keys:
        inputs_set = feature_sets[set].values
        dim[set] = inputs_set.shape[1]
        dimsum += dim[set]
        inputs = np.hstack((inputs, inputs_set))
        feature_names.extend(feature_sets[set].columns.tolist())
    scores = model.predict(inputs)
    scores = [(s * 100).round(2) for s in scores]
    return scores


def deepcpf1(cas12_sites, models_dir):
    '''
    Cpf1(cas12) on-target score "kinda"
    predicts the likelihood of getting a cas12 indel at the desired target
    This script is copied and modified from https://github.com/MyungjaeSong/Paired-Library
    Kim, H., Song, M., Lee, J. et al. In vivo high-throughput profiling of CRISPR–Cpf1 activity. Nat Methods 14, 153–159 (2017)
    '''
    model1,model2 = load_model_params('deepcpf1', models_dir)
    data_n = len(cas12_sites)
    #if chromatin_flag != 'ignore':
     #   CA = np.zeros((data_n, 1), dtype=int)

    SEQ = np.zeros((data_n, 34, 4), dtype=int)

    for l in range(data_n):
        seq = cas12_sites[l]
        for i in range(34):
            if seq[i] in "Aa":
                SEQ[l, i, 0] = 1
            elif seq[i] in "Cc":
                SEQ[l, i, 1] = 1
            elif seq[i] in "Gg":
                SEQ[l, i, 2] = 1
            elif seq[i] in "Tt":
                SEQ[l, i, 3] = 1
        #CA[l - 1, 0] = int(chromatin_accessibility) * 100

    scores = model2.predict([SEQ], batch_size=50, verbose=0).tolist()

    return scores


def deepABE(abesites):
    '''
    ABE Efficiency Scoring
    This script is copied and modified from https://github.com/MyungjaeSong/Paired-Library
    Zhang, C., Yang, Y., Qi, T. et al. Prediction of base editor off-targets by deep learning.
    Nat Commun 14, 5358 (2023). https://doi.org/10.1038/s41467-023-41004-3
    30bp input 30 bp target sequence (4 bp + 20 bp protospacer + PAM + 3 bp)
    '''
    pass
    #model = load_model_params(score_name='deepabe')


# === Microhomology Scoring ===
def oofscore(seq):
    '''
    copied and adapted code from Bae et al. https://www.nature.com/articles/nmeth.3015
    computes both microhomology and out-of-frame score
    A measurement of how likely an out-of-frame deletion occurs after a knock-out experiment
    based on microhomology
    scoring range 0-100. The higher the oof score, the more deletions have a length that is not a multiple of three
     A score above 66 is recommended
    The higher the oof score, the more deletions have a length that is not a multiple of three
    '''
    length_weight = 20.0
    left = 30
    right = len(seq) - int(left)

    s1 = []
    for k in range(2, left)[::-1]:
        for j in range(left, left + right - k + 1):
            for i in range(0, left - k + 1):
                if seq[i:i + k] == seq[j:j + k]:
                    length = j - i
                    s1.append(seq[i:i + k] + '\t' + str(i) + '\t' + str(i + k) + '\t' + str(j) + '\t' + str(j + k) + '\t' + str(length))

    if s1 != "":
        list_f1 = s1
        sum_score_3 = 0
        sum_score_not_3 = 0

        for i in range(len(list_f1)):
            n = 0
            score_3 = 0
            score_not_3 = 0
            line = list_f1[i].split('\t')
            scrap = line[0]
            left_start = int(line[1])
            left_end = int(line[2])
            right_start = int(line[3])
            right_end = int(line[4])
            length = int(line[5])

            for j in range(i):
                line_ref = list_f1[j].split('\t')
                left_start_ref = int(line_ref[1])
                left_end_ref = int(line_ref[2])
                right_start_ref = int(line_ref[3])
                right_end_ref = int(line_ref[4])

                if (left_start >= left_start_ref) and (left_end <= left_end_ref) and (
                        right_start >= right_start_ref) and (right_end <= right_end_ref):
                    if (left_start - left_start_ref) == (right_start - right_start_ref) and (
                            left_end - left_end_ref) == (right_end - right_end_ref):
                        n += 1
                else:
                    pass

            if n == 0:
                if (length % 3) == 0:
                    length_factor = round(1 / exp((length) / (length_weight)), 3)
                    num_GC = len(re.findall('G', scrap)) + len(re.findall('C', scrap))
                    score_3 = 100 * length_factor * ((len(scrap) - num_GC) + (num_GC * 2))

                elif (length % 3) != 0:
                    length_factor = round(1 / exp((length) / (length_weight)), 3)
                    num_GC = len(re.findall('G', scrap)) + len(re.findall('C', scrap))
                    score_not_3 = 100 * length_factor * ((len(scrap) - num_GC) + (num_GC * 2))
            sum_score_3 += score_3
            sum_score_not_3 += score_not_3

        mh_score = round(sum_score_3 + sum_score_not_3,2)
        oof_score = round((sum_score_not_3) * 100 / (sum_score_3 + sum_score_not_3),2)

    return mh_score, oof_score


# === Off-target Specifity Scoring ===
def cfd_score(seq1, seq2, models_dir):
    '''
    Doench 2016 off-target scoring
    Doench, Fusi, et al.  Nature Biotechnology 34, 184–191 (2016)."

    '''
    mm_scores, pam_scores = load_model_params('cfd', models_dir)
    pam = seq2[-3:]
    seq1 = seq1.upper().replace('T', 'U')
    seq2 = seq2[:-3].upper().replace('T', 'U')
    m_seq1 = re.search('[^ATCGU\\-]', seq1)
    m_seq2 = re.search('[^ATCGU\\-]', seq2)

    score =1

    if (m_seq1 is None) and (m_seq2 is None):
        if seq1 != seq2:
            shorter, longer = sorted([seq1, seq2], key=len)
            for i in range(-len(shorter), 0):
                if (seq1[i] != seq2[i]):
                    key = 'r' + seq1[i] + ':d' + revcom(seq2[i]) + ',' + str(20 + i + 1)
                    score *= mm_scores[key]

            score *= pam_scores[pam[-2:]]
    else:
        score = -1
    return round(score,4)


def cfd_spec_score(sum_cfd_scores):
    '''
    on_target_seq is spacer site and off_target includes pam
    '''

    guide_cfd_score = 100 / (100+sum_cfd_scores)
    guide_cfd_score = round(guide_cfd_score*100,2)
    return guide_cfd_score


'''
Test
pam = 'AGG'
spacer = 'GTGCGGCTGGCCCAGGACCT'
target = spacer + pam
target30mer = 'CTTGTGCGGCTGGCCCAGGACCTAGGCGAG'
target60mer = 'ATCTCTTACAACGACTTCTTGTGCGGCTGGCCCAGGACCTAGGCGAGGCAGTAGGGGATGACA'
off_target_seqs = ['GTGGGGCTGACCCAGGACCTGAG','GGGCCTCTGGCCCAGGACCTGGG']


print('Microhomology Out-Of-Frame score: ',oofscore(seq = target60mer))
# Ans: 3730.1, 64.2

print('Azimuth on-target score: ',azimuth([target30mer])[0])
#Ans: 0.38

print('CFD score ',cfd_score(spacer, off_target_seqs[1]))
print('CFD score ',cfd_score(spacer, off_target_seqs[0]))
#0.44, 0.12

print('CFD spec score ',cfd_spec_score(sum([0.44,0.12])))
#99

seq1, seq2 = 'GTGCGGCTGGCCCAGGACC-T', 'GTGCtGCTacCCCAGGACCCTCGG'
print('CFD score ',cfd_score(seq1, seq2))

DeepCpf1
cas12_sites = ['TGACTTTGAATGGAGTCGTGAGCGCAAGAACGCT','GTTATTTGAGCAATGCCACTTAATAAACATGTAA']
#= [55.437206] [78.50043]
'''



# This file is copied and modified from https://github.com/MicrosoftResearch/Azimuth
# This code was modified to use with python 3 and needed for percent and position scoring
# was removed to simplify code
# Azimuth license applies to this file

# Native Modules
import itertools
# Installed Modules
import pandas as pd
import numpy as np

import Bio.SeqUtils.MeltingTemp as Tm


def countGC(s, length_audit=True):
    '''
    GC content for only the 20mer, as per the Doench paper/code
    '''
    if length_audit:
        assert len(s) == 30, "seems to assume 30mer"
    return len(s[4:24].replace('A', '').replace('T', ''))


def nucleotide_features(s, order, include_pos_independent, max_index_to_use, prefix="", feature_type='all',
                        raw_alphabet=['A', 'T', 'C', 'G']):
    '''
    compute position-specific order-mer features for the 4-letter alphabet
    (e.g. for a sequence of length 30, there are 30*4 single nucleotide features
          and (30-1)*4^2=464 double nucleotide features
    '''
    assert feature_type in ['all', 'pos_independent', 'pos_dependent']
    if max_index_to_use <= len(s):
        # print "WARNING: trimming max_index_to use down to length of string=%s" % len(s)
        max_index_to_use = len(s)
    if max_index_to_use is not None:
        s = s[:max_index_to_use]
    # assert(len(s)==30, "length not 30")
    # s = s[:30] #cut-off at thirty to clean up extra data that they accidentally left in, and were instructed to ignore in this way
    alphabet = get_alphabet(order, raw_alphabet=raw_alphabet)
    features_pos_dependent = np.zeros(len(alphabet) * (len(s) - (order - 1)))
    features_pos_independent = np.zeros(np.power(len(raw_alphabet), order))

    index_dependent = []
    index_independent = []

    for position in range(0, len(s) - order + 1, 1):
        for l in alphabet:
            index_dependent.append('%s%s_%d' % (prefix, l, position))

    for l in alphabet:
        index_independent.append('%s%s' % (prefix, l))

    for position in range(0, len(s) - order + 1, 1):
        nucl = s[position:position + order]
        features_pos_dependent[alphabet.index(nucl) + (position * len(alphabet))] = 1.0
        features_pos_independent[alphabet.index(nucl)] += 1.0

        # this is to check that the labels in the pd df actually match the nucl and position
        assert index_dependent[alphabet.index(nucl) + (position * len(alphabet))] == '%s%s_%d' % (
        prefix, nucl, position)
        assert index_independent[alphabet.index(nucl)] == '%s%s' % (prefix, nucl)

    # index_independent = ['%s_pi.Order%d_P%d' % (prefix, order,i) for i in range(len(features_pos_independent))]
    # index_dependent = ['%s_pd.Order%d_P%d' % (prefix, order, i) for i in range(len(features_pos_dependent))]

    if np.any(np.isnan(features_pos_dependent)):
        raise Exception("found nan features in features_pos_dependent")
    if np.any(np.isnan(features_pos_independent)):
        raise Exception("found nan features in features_pos_independent")

    if feature_type == 'all' or feature_type == 'pos_independent':
        if feature_type == 'all':
            res = pd.Series(features_pos_dependent, index=index_dependent), pd.Series(features_pos_independent,
                                                                                      index=index_independent)
            assert not np.any(np.isnan(res.values))
            return res
        else:
            res = pd.Series(features_pos_independent, index=index_independent)
            assert not np.any(np.isnan(res.values))
            return res

    res = pd.Series(features_pos_dependent, index=index_dependent)
    assert not np.any(np.isnan(res.values))
    return res


def Tm_feature(data, pam_audit=True, learn_options=None):
    '''
    assuming '30-mer'is a key
    get melting temperature features from:
        0-the 30-mer ("global Tm")
        1-the Tm (melting temperature) of the DNA:RNA hybrid from positions 16 - 20 of the sgRNA, i.e. the 5nts immediately proximal of the NGG PAM
        2-the Tm of the DNA:RNA hybrid from position 8 - 15 (i.e. 8 nt)
        3-the Tm of the DNA:RNA hybrid from position 3 - 7  (i.e. 5 nt)
    '''
    if learn_options is None or 'Tm segments' not in learn_options.keys():
        segments = [(19, 24), (11, 19), (6, 11)]
    else:
        segments = learn_options['Tm segments']
    sequence = data['30mer'].values
    featarray = np.ones((sequence.shape[0], 4))
    for i, seq in enumerate(sequence):
        if seq[25:27] != "GG":
            raise Exception("expected GG but found %s" % seq[25:27])
        rna = False
        featarray[i, 0] = Tm.Tm_NN(seq)  # 30mer Tm
        featarray[i, 1] = Tm.Tm_NN(seq[segments[0][0]:segments[0][1]], )  # 5nts immediately proximal of the NGG PAM
        featarray[i, 2] = Tm.Tm_NN(seq[segments[1][0]:segments[1][1]])  # 8-mer
        featarray[i, 3] = Tm.Tm_NN(seq[segments[2][0]:segments[2][1]])  # 5-mer

    feat = pd.DataFrame(featarray, index=data.index,
                        columns=["Tm global_%s" % rna, "5mer_end_%s" % rna, "8mer_middle_%s" % rna,
                                 "5mer_start_%s" % rna])

    return feat


def get_alphabet(order, raw_alphabet=['A', 'T', 'C', 'G']):
    alphabet = ["".join(i) for i in itertools.product(raw_alphabet, repeat=order)]
    return alphabet


def NGGX_interaction_feature(data, pam_audit=True):
    '''
    assuming 30-mer, grab the NGGX _ _ positions, and make a one-hot
    encoding of the NX nucleotides yielding 4x4=16 features
    '''
    sequence = data['30mer'].values
    feat_NX = pd.DataFrame()
    # check that GG is where we think
    for seq in sequence:
        if seq[25:27] != "GG":
            raise Exception("expected GG but found %s" % seq[25:27])
        NX = seq[24] + seq[27]
        NX_onehot = nucleotide_features(NX, order=2, feature_type='pos_dependent', max_index_to_use=2, prefix="NGGX")
        # NX_onehot[:] = np.random.rand(NX_onehot.shape[0]) ##TESTING RANDOM FEATURE
        feat_NX = pd.concat([feat_NX, NX_onehot], axis=1)
    return feat_NX.T


def apply_nucleotide_features(seq_data_frame, order, num_proc, include_pos_independent, max_index_to_use, prefix=""):
    if include_pos_independent:
        feat_pd = seq_data_frame.apply(nucleotide_features, args=(order, max_index_to_use, prefix, 'pos_dependent'))
        feat_pi = seq_data_frame.apply(nucleotide_features, args=(order, max_index_to_use, prefix, 'pos_independent'))
        assert not np.any(np.isnan(feat_pd)), "nans here can arise from sequences of different lengths"
        assert not np.any(np.isnan(feat_pi)), "nans here can arise from sequences of different lengths"
        return feat_pd, feat_pi
    else:
        feat_pd = seq_data_frame.apply(nucleotide_features, args=(order, max_index_to_use, prefix, 'pos_dependent'))
        assert not np.any(np.isnan(feat_pd)), "found nan in feat_pd"
        return feat_pd


def nucleotide_features(s, order, max_index_to_use, prefix="", feature_type='all', raw_alphabet=['A', 'T', 'C', 'G']):
    '''
    compute position-specific order-mer features for the 4-letter alphabet
    (e.g. for a sequence of length 30, there are 30*4 single nucleotide features
          and (30-1)*4^2=464 double nucleotide features
    '''
    assert feature_type in ['all', 'pos_independent', 'pos_dependent']
    if max_index_to_use <= len(s):
        # print "WARNING: trimming max_index_to use down to length of string=%s" % len(s)
        max_index_to_use = len(s)

    if max_index_to_use is not None:
        s = s[:max_index_to_use]
    # assert(len(s)==30, "length not 30")
    # s = s[:30] #cut-off at thirty to clean up extra data that they accidentally left in, and were instructed to ignore in this way
    alphabet = get_alphabet(order, raw_alphabet=raw_alphabet)
    features_pos_dependent = np.zeros(len(alphabet) * (len(s) - (order - 1)))
    features_pos_independent = np.zeros(np.power(len(raw_alphabet), order))

    index_dependent = []
    index_independent = []

    for position in range(0, len(s) - order + 1, 1):
        for l in alphabet:
            index_dependent.append('%s%s_%d' % (prefix, l, position))

    for l in alphabet:
        index_independent.append('%s%s' % (prefix, l))

    for position in range(0, len(s) - order + 1, 1):
        nucl = s[position:position + order]
        features_pos_dependent[alphabet.index(nucl) + (position * len(alphabet))] = 1.0
        features_pos_independent[alphabet.index(nucl)] += 1.0

        # this is to check that the labels in the pd df actually match the nucl and position
        assert index_dependent[alphabet.index(nucl) + (position * len(alphabet))] == '%s%s_%d' % (
        prefix, nucl, position)
        assert index_independent[alphabet.index(nucl)] == '%s%s' % (prefix, nucl)

    if np.any(np.isnan(features_pos_dependent)):
        raise Exception("found nan features in features_pos_dependent")
    if np.any(np.isnan(features_pos_independent)):
        raise Exception("found nan features in features_pos_independent")

    if feature_type == 'all' or feature_type == 'pos_independent':
        if feature_type == 'all':
            res = pd.Series(features_pos_dependent, index=index_dependent), pd.Series(features_pos_independent,
                                                                                      index=index_independent)
            assert not np.any(np.isnan(res.values))
            return res
        else:
            res = pd.Series(features_pos_independent, index=index_independent)
            assert not np.any(np.isnan(res.values))
            return res

    res = pd.Series(features_pos_dependent, index=index_dependent)
    assert not np.any(np.isnan(res.values))
    return res

def check_feature_set(feature_sets):
    '''
    Ensure the # of people is the same in each feature set
    '''
    assert feature_sets != {}, "no feature sets present"

    N = None
    for ft in feature_sets.keys():
        N2 = feature_sets[ft].shape[0]
        if N is None:
            N = N2
        else:
            assert N >= 1, "should be at least one individual"
            assert N == N2, "# of individuals do not match up across feature sets"

    for set in feature_sets.keys():
        if np.any(np.isnan(feature_sets[set])):
            raise Exception("found Nan in set %s" % set)


def get_all_order_nuc_features(data, feature_sets, learn_options, maxorder, max_index_to_use, prefix="", quiet=False):
    for order in range(1, maxorder + 1):
        nuc_features_pd, nuc_features_pi = apply_nucleotide_features(data, order, learn_options["num_proc"],
                                                                     include_pos_independent=True,
                                                                     max_index_to_use=max_index_to_use, prefix=prefix)
        feature_sets['%s_nuc_pd_Order%i' % (prefix, order)] = nuc_features_pd
        if learn_options['include_pi_nuc_feat']:
            feature_sets['%s_nuc_pi_Order%i' % (prefix, order)] = nuc_features_pi
        check_feature_set(feature_sets)



def gc_features(data, audit=True):
    gc_count = data['30mer'].apply(lambda seq: countGC(seq, audit))
    gc_count.name = 'GC count'
    gc_above_10 = (gc_count > 10) * 1
    gc_above_10.name = 'GC > 10'
    gc_below_10 = (gc_count < 10) * 1
    gc_below_10.name = 'GC < 10'
    return gc_above_10, gc_below_10, gc_count


def normalize_features(data, axis):
    '''
    input: Pandas.DataFrame of dtype=np.float64 array, of dimensions
    mean-center, and unit variance each feature
    '''
    data -= data.mean(axis)
    data /= data.std(axis)
    # remove rows with NaNs
    data = data.dropna(1)
    if np.any(np.isnan(data.values)): raise Exception("found NaN in normalized features")
    return data


def normalize_feature_sets(feature_sets):
    '''
    zero-mean, unit-variance each feature within each set
    '''

    print("Normalizing features...")
    new_feature_sets = {}
    for set in feature_sets:
        new_feature_sets[set] = normalize_features(feature_sets[set], axis=0)
        if np.any(np.isnan(new_feature_sets[set].values)):
            raise Exception("found Nan feature values in set=%s" % set)
        assert new_feature_sets[set].shape[1] > 0, "0 columns of features"
    return new_feature_sets


def featurize_data(data, learn_options, Y, pam_audit=True, length_audit=True):
    '''
    data = Xdf
    assumes that data contains the 30mer
    returns set of features from which one can make a kernel for each one
    '''
    all_lens = data['30mer'].apply(len).values
    unique_lengths = np.unique(all_lens)
    num_lengths = len(unique_lengths)
    assert num_lengths == 1, "should only have sequences of a single length, but found %s: %s" % (
    num_lengths, str(unique_lengths))

    feature_sets = {}

    if learn_options["nuc_features"]:
        # spectrum kernels (position-independent) and weighted degree kernels (position-dependent)
        get_all_order_nuc_features(data['30mer'], feature_sets, learn_options, learn_options["order"],
                                   max_index_to_use=30, quiet=False)

    check_feature_set(feature_sets)

    if learn_options["gc_features"]:
        gc_above_10, gc_below_10, gc_count = gc_features(data, length_audit)
        feature_sets['gc_above_10'] = pd.DataFrame(gc_above_10)
        feature_sets['gc_below_10'] = pd.DataFrame(gc_below_10)
        feature_sets['gc_count'] = pd.DataFrame(gc_count)

    if learn_options["include_NGGX_interaction"]:
        feature_sets["NGGX"] = NGGX_interaction_feature(data)

    if learn_options["include_Tm"]:
        feature_sets["Tm"] = Tm_feature(data, pam_audit, learn_options=None)


    check_feature_set(feature_sets)

    reordered_feature_sets = {}
    keys = ['gc_count', '_nuc_pd_Order2', '_nuc_pd_Order1', 'gc_above_10', '_nuc_pi_Order1', '_nuc_pi_Order2', 'Tm',
            'gc_below_10', 'NGGX']
    for k in keys:
        reordered_feature_sets[k] = feature_sets[k]

    return reordered_feature_sets






# Native modules
import os
import copy
import pandas as pd
from pathlib import Path
# Installed modules
import urllib.error
import urllib3
import xmltodict
# Biopython
from Bio import Entrez
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

# Load config file
# with open("config/guide_prediction_default_template.yaml", "r") as f:
# 	config = yaml.load(f, Loader=yaml.FullLoader)


def list_lines_from_txt(file_path):
	with open(file_path, "r") as file_handle:
		list_of_entries = []
		for line in file_handle:
			clean_line = line.strip()
			list_of_entries.append(clean_line)
	return list_of_entries


def get_clinvar_uid(word_query):
	handle = Entrez.esearch(db="clinvar", term=f"{word_query}", idtype="acc")
	search_record = Entrez.read(handle)
	try:
		uid = search_record['IdList'][0]
	except IndexError:
		raise "No HGVSs were found for this search"
	handle.close()
	return uid


def elink_routine(dbfrom, dbto, hit_uid):
	dup_check = []
	not_found = ""
	linked = ""
	link_record = ""
	server_attempts = 0
	try:
		handle = Entrez.elink(dbfrom=dbfrom, db=dbto, id=f"{hit_uid}")
	except urllib.error.HTTPError as err:
		if err.code == 500:
			print(f"An internal server error occurred while handling the accession {hit_uid}")
			not_found = hit_uid
			return linked, hit_uid, not_found
	try:
		link_record = Entrez.read(handle)
	except RuntimeError:
		not_found = hit_uid
	if link_record:
		try:
			linked = link_record[0]['LinkSetDb'][0]['Link'][0]['Id']
			if linked not in dup_check:
				dup_check.append(linked)
		except (IndexError, KeyError):
			not_found = hit_uid
	handle.close()
	return linked, hit_uid, not_found


def url2xml_dict(url):
	http = urllib3.PoolManager()
	file = http.request('GET', str(url))
	data = file.data
	data = xmltodict.parse(data)
	return data


def get_gene_xml(query):
	# Get XML records from Entrez' Gene database
	handle = Entrez.efetch(db="gene", id=f"{query}", rettype="xml", retmode="text")
	xml_data = url2xml_dict(handle.url)
		# Returns a list  of Genbank SeqRecords objects
	return xml_data


def get_gene_coords(gene_xml_dict):
	gene_coords = {}
	chromosomes = gene_xml_dict['Entrezgene-Set']['Entrezgene']['Entrezgene_locus']['Gene-commentary']
	chromosomes_list_length = len(gene_xml_dict['Entrezgene-Set']['Entrezgene']['Entrezgene_locus']['Gene-commentary'])
	gene_uid = gene_xml_dict['Entrezgene-Set']['Entrezgene']['Entrezgene_track-info']['Gene-track']['Gene-track_geneid']

	# Get gene coordinates for each individual chromosome
	for chr_index in range(0, chromosomes_list_length):
		chromosome_acc = chromosomes[chr_index]['Gene-commentary_accession']
		chromosome_version = chromosomes[chr_index]['Gene-commentary_version']
		chromosome_id = f"{chromosome_acc}.{chromosome_version}"
		gene_coords.setdefault(gene_uid, {}).setdefault(chromosome_id, [
			gene_xml_dict['Entrezgene-Set']['Entrezgene']['Entrezgene_locus']['Gene-commentary'][chr_index]['Gene-commentary_seqs']['Seq-loc']['Seq-loc_int']['Seq-interval']['Seq-interval_from'],
			gene_xml_dict['Entrezgene-Set']['Entrezgene']['Entrezgene_locus']['Gene-commentary'][chr_index]['Gene-commentary_seqs']['Seq-loc']['Seq-loc_int']['Seq-interval']['Seq-interval_to']
		])
	return gene_coords


def get_genbank(gene_coords, win_size):
	prot_dict = {}
	record_target = {}
	print("Fetch GenBank data via Entrez")
	for gene_uid in gene_coords:
		for chromosome_id in gene_coords[gene_uid]:
			print("Processing Gene gi", gene_uid, " from chromosome ", chromosome_id)
			# Fetch Genbank entry
			handle = Entrez.efetch(db="nucleotide", id=str(chromosome_id), rettype="gbwithparts", retmode="text")
			record = SeqIO.read(handle, "genbank")
			# Select the 'features' object from the GB file
			features = record.features[0]

			#
			# prot_id = qualifiers["protein_id"][0]
			# # Search for the protein-ids of interest
			# if re.search(prot_id, uid_to_acc[query][hit_uid][0]):

			# Process feature information for future ref
			f_start = int(gene_coords[gene_uid][chromosome_id][0])
			f_end = int(gene_coords[gene_uid][chromosome_id][1])
			f_strand = features.strand
			highlight_feature = copy.deepcopy(features)
			highlight_feature.type = "highlight"
			# Set start/end coords using window size
			start = max(int(min([f_start, f_end])) - win_size, 0)
			end = min(int(max([f_start, f_end])) + win_size + 1, len(record.seq))
			f_len = end - start

			# Create a SeqRecord object with the feature of interest
			record_focused = SeqRecord(
				id=record.id,
				annotations=record.annotations,
				dbxrefs=record.dbxrefs,
				seq=record.seq[start:end + 1],
				description=record.description
			)
			record_focused.features.append(highlight_feature)

			# Create a fasta record with the feature of interes
			fasta_focused: record.seq[start:end + 1]

			# Gather protein data for reference
			prep_prot_dict = {
			                  "nuccore_acc": record.id,
			                  # "region_seq": record.seq[start:end + 1],
			                  "window_start": start,
			                  "window_end": end,
			                  "feature_start": f_start,
			                  "feature_end": f_end,
			                  "strand": f_strand,
			                  "feature_len": f_len,
			                  }
			prot_dict.setdefault(gene_uid, {}).setdefault(record.id,  prep_prot_dict)
			record_target.setdefault(gene_uid, {}).setdefault(record.id, record_focused)
	return prot_dict, record_target


def create_new_dir(path):
	Path(path).mkdir(parents=True, exist_ok=True)


def export_gbs(gb_dict, hgvs_id, parent_path):
	for gene_uid in gb_dict:
		hgvs_output_directory = f"{parent_path}{os.sep}{gene_uid}:{hgvs_id}"
		genbank_output_directory = f"{hgvs_output_directory}{os.sep}gb"
		fasta_output_directory = f"{hgvs_output_directory}{os.sep}fna"
		create_new_dir(genbank_output_directory)
		create_new_dir(fasta_output_directory)
		for chromosome_id in gb_dict[gene_uid]:
			gbk = gb_dict[gene_uid][chromosome_id]
			with open(f"{genbank_output_directory}{os.sep}{chromosome_id}.gb", "w") as gb_handle:
				SeqIO.write(gbk, gb_handle, "genbank")
			with open(f"{fasta_output_directory}{os.sep}{chromosome_id}.fna", "w") as gb_handle:
				SeqIO.write(gbk, gb_handle, "fasta")


# DEBUG INPUT
# hgvs_id = 'NM_001355224.2:c.867C>A'
# entrez_login = 'thedoudnalab@gmail.com'
# genomic_window_size = 1000

def main():
	# SNAKEMAKE IMPORTS
	# Inputs
	hgvs_file_path = str(snakemake.input.hgvs_list)
	# Outputs
	report_out = str(snakemake.output.database_report)
	# Params
	genomic_window_size = int(snakemake.params.window_size)
	entrez_login = str(snakemake.params.entrez_login)
	# Wildcards
	output_directory = str(snakemake.wildcards.output_directory)
	run_name = str(snakemake.wildcards.run)

	# Get list of HGVSs from file
	hgvs_list = list_lines_from_txt(hgvs_file_path)

	# Set the analysis directory name
	analysis_directory = f"{output_directory}{os.sep}{run_name}"

	for hgvs_id in hgvs_list:
		# Entrez login
		Entrez.email = entrez_login

		# Find NCBI's UID for a given HGVS query
		clinvar_uid = get_clinvar_uid(hgvs_id)

		# Retrieve GENE object from Entrez based on clinvar
		(linked, hit_id, notfound) = elink_routine('clinvar', 'gene', clinvar_uid)

		# Generate a genomic coordinate dictionary from the GENE object
		ncbi_gene_dict = get_gene_xml(linked)
		gene_coordinates = get_gene_coords(ncbi_gene_dict)

		# Consolidate focused GenBank entries
		(report_dict, focus_gb_dict) = get_genbank(gene_coordinates, genomic_window_size)

		# Export focused GenBank to text files
		export_gbs(focus_gb_dict, hgvs_id, analysis_directory)

		# Export report file
		pd.DataFrame.from_dict(report_dict[list(report_dict.keys())[0]]).to_csv(report_out)


if __name__ == "__main__":
	main()



# == Native Modules ==
import subprocess
import gzip
import shutil
from datetime import datetime
import secrets
import string
import pytz
import os
import os.path
import re
import pathlib
import pickle
# == Installed Modules ==
from alive_progress import alive_bar
import yaml
from importlib_resources import files
from Bio import SeqIO
import boto3
from botocore.exceptions import NoCredentialsError
# == Project Modules ==


def compress_file(file_path: str):
	if not is_gzipped(file_path):
		# If not gzipped, compress the file
		with open(file_path, 'rb') as f_in, gzip.open(file_path + '.gz', 'wb') as f_out:
			shutil.copyfileobj(f_in, f_out)
		print(f"File '{file_path}' compressed successfully.")
	if is_gzipped(file_path):
		cmd_rename = f"mv {file_path} {file_path}.gz"
		subprocess.run(cmd_rename, shell=True)
		print("This file is already compressed.")
		print(f"Created a copy of the file input on: {file_path}.gz")


def consolidate_s3_download(content, parent_folder):
	consolidated_downloadable = []
	for content_idx in range(0, len(content)):
		key = content[content_idx]['Key']
		if key != f"{parent_folder}/":
			consolidated_downloadable.append(content[content_idx])
	return consolidated_downloadable


def date_tag():
	# Create a random string of 20 characters
	random_str = ''.join(secrets.choice(string.ascii_letters + string.digits) for _ in range(10))

	# Set the timezone to PST
	pst = pytz.timezone('America/Los_Angeles')
	# Get the current date and time
	current_datetime = datetime.now(pst)
	# Format the date as a string with day, hour, minute, and second
	formatted_date = f"{current_datetime.strftime('%y%m%d%H%M%S%f')}_{random_str}"

	return formatted_date


def download_s3_objects(s3_bucket_name: str, s3_object_name: str, destination_path: str):
	"""
	Downloads a file or every file in a given AWS S3 bucket
	:param s3_bucket_name:
	:param s3_object_name:
	:param destination_path:
	:return:
	"""
	s3 = boto3.client('s3')
	try:
		response = s3.list_objects_v2(Bucket=s3_bucket_name, Prefix=s3_object_name)
		content = response.get('Contents', [])
		clean_content = consolidate_s3_download(content, s3_object_name)  # Skip the S3 object that denotes the parent folder
	except NoCredentialsError:
		prRed("AWS Credentials not available. Please sign in then try again!")
		exit(0)
	with alive_bar(len(clean_content), title=f'Downloading s3://{s3_bucket_name}/{s3_object_name}') as bar:
		for content_idx in range(0, len(clean_content)):
			key = clean_content[content_idx]['Key']
			source_file = key.split("/")[-1]
			destination_file = os.path.join(destination_path, source_file)
			if file_exists(destination_file):
				print(f"Skipping existing file: {destination_file}")
				bar()
				continue
			pathlib.Path(destination_path).mkdir(parents=True, exist_ok=True)
			s3.download_file(s3_bucket_name, key, destination_file)
			bar()


def file_exists(file_path):
	return os.path.exists(file_path)


def handle_shell_exception(subprocess_result, shell_command, verbose: bool):
	# === Handle SMK exceptions through subprocess
	#   == Unlock directory if necessary for SMK run
	if re.findall("Directory cannot be locked.", subprocess_result.stdout):
		print("--> Target directory locked. Unlocking...")
		unlock_smk_command = f"{shell_command} --unlock"
		launch_shell_cmd(unlock_smk_command, verbose)
		launch_shell_cmd(shell_command, verbose)
		return
	#   == Skipping rule call that has already been completed
	if re.findall(r"ValueError: min\(\) arg is an empty sequence", subprocess_result.stderr):
		print("--> A consensus FASTA has already been generated for this job. Skipping.")
		return
	if not re.findall(r"ValueError: min\(\) arg is an empty sequence", subprocess_result.stderr):
		if verbose:
			print(subprocess_result.stderr)
			prGreen(subprocess_result.stdout)
		return


def is_bgzipped(file_path: str) -> bool:
	with open(file_path, 'rb') as f:
		# Check if the file starts with the bgzip magic bytes
		# BGZF magic bytes: 1f 8b 08 04
		return f.read(4) == b'\x1f\x8b\x08\x04'


def is_gzipped(file_path: str):
	with open(file_path, 'rb') as f:
		# Check if the file starts with the gzip magic bytes
		return f.read(2) == b'\x1f\x8b'


def launch_shell_cmd(command: str, verbose=False, **kwargs):
	message = kwargs.get('message', False)
	check_exist = kwargs.get('check_exist', False)
	if message:
		verbose = False
		print(message)
	if check_exist:
		if os.path.isfile(check_exist):
			print(f"File {check_exist} exists. Skipping process.")
			return
	if verbose:
		prCyan(f"--> Invoking command-line call:\n{command}")

	result = subprocess.run(command,
	                        shell=True,
	                        stderr=subprocess.PIPE,
	                        stdout=subprocess.PIPE,
	                        universal_newlines=True
	                        )
	handle_shell_exception(result, command, verbose)


def list_files_by_extension(root_path, extension: str):
	file_list = []
	for root, dirs, files in os.walk(root_path, topdown=False):
		for name in files:
			if name.endswith(extension):
				file_list.append(os.path.join(root, name))
	return file_list


def parse_editor_request(request):
	processed_request = request
	if type(processed_request) is list:
		processed_request = request.split(',')

	return processed_request


def pickle_chromosomes(genome_fasta, output_dir):
	records = SeqIO.parse(open(genome_fasta, 'rt'), "fasta")
	with alive_bar(25, title=f'Serializing human chromosomes') as bar:
		for record in records:
			if re.search(r"chr\w{0,2}$", record.id, re.IGNORECASE):
				outfile = f"{output_dir}/{record.id}.pkl"
				with open(outfile, 'ab') as gfile:
					pickle.dump(record, gfile)
					bar()


def prCyan(skk):
	print("\033[96m {}\033[00m".format(skk))


def prGreen(skk):
	print("\033[92m {}\033[00m".format(skk))


def prRed(skk):
	print("\033[0;31;47m {}\033[00m".format(skk))


def project_file_path(path_from_toplevel: str, filename: str):
	"""
	There are two top-level directories in the current version of mEdit: snakemake and config
	From either of these paths, the respective *.smk and *.yaml files can be accessed
	:param path_from_toplevel:
	:param filename:
	:return:
	"""
	return str(files(path_from_toplevel).joinpath(filename))


def set_export(outdir: str):
	if os.path.exists(outdir):
		print(f'--> Skipping directory creation: {outdir}')
	# Create outdir only if it doesn't exist
	if not os.path.exists(outdir):
		print(f'Directory created on: {outdir}')
		pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)
	return outdir


def write_yaml_to_file(py_obj, filename: str):
	with open(f'{filename}', 'w', ) as f:
		yaml.safe_dump(py_obj, f, sort_keys=False, default_style='"')
	print(f'--> Configuration file created: {filename}')



# == Native Modules ==
from os.path import abspath
# == Installed Modules ==
import yaml
# == Project Modules ==
from prog.medit_lib import (compress_file,
							download_s3_objects,
							pickle_chromosomes,
							project_file_path,
							launch_shell_cmd,
							set_export,
							write_yaml_to_file)


def dbset(args):
	# === Load template configuration file ===
	config_path = project_file_path("smk.config", "medit_database.yaml")
	with open(config_path, 'r') as config_handle:
		config_db_template = yaml.safe_load(config_handle)

	# === Load Database Path ===
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	config_db_dir_path = f"{db_path_full}/config_db"
	# == Load args
	threads = args.threads
	latest_genome_download = args.latest_reference
	custom_reference = args.custom_reference

	vcf_dir_path = f"{db_path_full}/standard/source_vcfs"
	config_db_path = f"{config_db_dir_path}/config_db.yaml"

	set_export(vcf_dir_path)
	set_export(config_db_dir_path)
	# === Assign Variables to Configuration File ===
	#   == Parent Database Path
	config_db_template['meditdb_path'] = f"{db_path_full}"
	#   == Assign jobtag and Fasta root path ==
	fasta_root_path = f"{db_path_full}/{config_db_template['fasta_root_path']}"
	config_db_template['fasta_root_path'] = fasta_root_path
	#   == Assign Editor pickles path ==
	config_db_template["editors"] = f"{db_path_full}/{config_db_template['editors']}"
	config_db_template["base_editors"] = f"{db_path_full}/{config_db_template['base_editors']}"
	config_db_template["models_path"] = f"{db_path_full}/{config_db_template['models_path']}"
	#   == Parse the Processed Tables folder and its contents ==
	processed_tables = f"{db_path_full}/{config_db_template['processed_tables']}"
	config_db_template["processed_tables"] = f"{processed_tables}"
	# config_db_template["simple_tables"] = f"{processed_tables}/{config_db_template['simple_tables']}"
	# config_db_template["hgvs_lookup"] = f"{processed_tables}/{config_db_template['hgvs_lookup']}"
	# config_db_template["clinvar_update"] = f"{processed_tables}/{config_db_template['clinvar_update']}"
	config_db_template["refseq_table"] = f"{processed_tables}/{config_db_template['refseq_table']}"

	#   == Parse the Raw Tables folder and its contents ==
	# raw_tables = f"{db_path_full}/{config_db_template['raw_tables']}"
	# config_db_template["raw_tables"] = f"{raw_tables}"
	# config_db_template["clinvar_summary"] = f"{raw_tables}/{config_db_template['clinvar_summary']}"
	# config_db_template["hpa"] = f"{raw_tables}/{config_db_template['hpa']}"
	# config_db_template["gencode"] = f"{raw_tables}/{config_db_template['gencode']}"

	# === Download Data ===
	#   == SeqRecord Pickles
	print("# ---*--- Database of Genomic References ---*---")
	download_s3_objects("medit.db", "genome_pkl", fasta_root_path)
	# == Download the latest human reference genome by request
	if latest_genome_download:
		download_s3_objects("medit.db", "latest_genome_ref", fasta_root_path)
		local_latest_ref_path = f"{fasta_root_path}/latest_hg38.fa"
		pickle_chromosomes(local_latest_ref_path, fasta_root_path)
		launch_shell_cmd(f"bgzip -c -@ {threads} {local_latest_ref_path} > {local_latest_ref_path}.gz",
						 message="Compressing human reference genome")
		launch_shell_cmd(f"rm {local_latest_ref_path}",
						 message="Cleaning up unused files")
		config_db_template["latest_reference"] = "True"
	if custom_reference:
		local_custom_ref_path = f"{fasta_root_path}/custom_reference.fa"
		launch_shell_cmd(f"cp {custom_reference} {local_custom_ref_path}",
						 message="Setting up custom human reference genome")
		pickle_chromosomes(local_custom_ref_path, fasta_root_path)
		launch_shell_cmd(f"bgzip -c -@ {threads} {local_custom_ref_path} > {local_custom_ref_path}.gz",
						 message="Compressing human reference genome")
		launch_shell_cmd(f"rm {local_custom_ref_path}",
						 message="Cleaning up unused files")
		config_db_template["custom_reference"] = "True"

	# === Write YAML configs to mEdit Root Directory ===
	write_yaml_to_file(config_db_template, config_db_path)

	#   == HPRC VCF files Setup
	download_s3_objects("medit.db", "hprc", vcf_dir_path)

	#   == Processed Tables and Raw Tables
	print("# ---*--- Pre-Processed Background Data Sets ---*---")
	download_s3_objects("medit.db", "processed_tables.tar.gz", db_path_full)
	# download_s3_objects("medit.db", "raw_tables.tar.gz", db_path_full)
	download_s3_objects("medit.db", "pkl.tar.gz", db_path_full)

	#   == Decompress tar.gz files in the database
	print("# ---*--- Unpacking Background Data ---*---")
	launch_shell_cmd(f"gzip -d {db_path_full}/processed_tables.tar.gz", verbose=False,
					 check_exist=f"{db_path_full}/processed_tables")
	launch_shell_cmd(f"gzip -d {db_path_full}/pkl.tar.gz", verbose=False,
					 check_exist=f"{db_path_full}/pkl")
	# launch_shell_cmd(f"tar -xf {db_path_full}/raw_tables.tar --directory={db_path_full}/ && "
	#                  f"rm {db_path_full}/raw_tables.tar", message="Unpacking background tables")
	launch_shell_cmd(f"tar -xf {db_path_full}/processed_tables.tar --directory={db_path_full}/ && "
	                 f"rm {db_path_full}/processed_tables.tar",
					 message="Unpacking background tables", check_exist=f"{db_path_full}/processed_tables")
	launch_shell_cmd(f"tar -xf {db_path_full}/pkl.tar --directory={db_path_full}/ && "
	                 f"rm {db_path_full}/pkl.tar", message="Unpacking models", check_exist=f"{db_path_full}/pkl")
	launch_shell_cmd(f"gzip -d {config_db_template['refseq_table']}.gz", verbose=False,
					 check_exist=f"{config_db_template['refseq_table']}")



# == Native Modules ==
import subprocess
from os.path import abspath
import sys
# == Installed Modules ==
import yaml
# == Project Modules ==
from prog.medit_lib import (compress_file,
                            file_exists,
                            launch_shell_cmd,
                            parse_editor_request,
                            project_file_path,
                            prRed,
                            set_export,
                            write_yaml_to_file
                            )


def guide_prediction(args, jobtag):
	# == Load Run Parameters values ==
	query_input = args.query_input
	user_jobtag = args.user_jobtag
	root_dir = abspath(args.output)
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	mode = args.mode
	private_genomes = args.private_genome
	qtype = args.qtype_request
	editor_request = parse_editor_request(args.editor_request)
	be_request = parse_editor_request(args.be_request)

	# == Load SLURM-related values ==
	ncores = args.ncores
	maxtime = args.maxtime
	parallel_processes = args.parallel_processes
	dry_run = args.dry_run

	# == Define dynamic SMK call variables ==
	allowed_rules = ['']
	cluster_smk_setup = ['']
	smk_verbosity = [True]
	smk_run_triggers = ''
	dryrun_setup = ''

	# ->=== INPUT CHECKS ===<-
	#   == Check the presence of private genome among the inputs ==
	if private_genomes:
		mode = 'private'
		private_genomes = args.private_genome.split(",")
	#   == Check the request to run on a cluster
	if parallel_processes:
		# --> Upon SLURM run request, the guide_prediction.smk is split in two separate runs
		# --> That's because samtool's conda package crashes on a libcrypto error when
		#       it's deployed by snakemake on a SLURM node
		cluster_smk_setup = ['', '--cluster "sbatch -t {cluster.time} -n {cluster.cores}" '
		                         '--cluster-config config/medit_cluster.yaml']
		allowed_rules = ['--until "consensus_fasta"', '']
		smk_verbosity = [False, True]
	#   == Check the dry run request
	if dry_run:
		dryrun_setup = '-n'
	if user_jobtag:
		smk_run_triggers = '--rerun-triggers "mtime"'

	# ->=== OUTPUT SETUP ===<-
	# == Set import paths tied to the SMK pipeline
	config_db_path = f"{db_path_full}/config_db/config_db.yaml"
	# == Set export paths tied to the SMK pipeline ==
	config_dir_path = f"{root_dir}/config"
	# == Set export paths for dynamic YAML files ==
	dynamic_config_path = f"{config_dir_path}/config_{jobtag}.yaml"
	dynamic_cluster_path = f"{config_dir_path}/cluster_{jobtag}.yaml"
	# == Create sub-folders to host VCFs, and config files ==
	set_export(config_dir_path)

	# ->=== CONFIG FILES IMPORT ===<-
	#   == Load template configuration files ==
	config_template_path = project_file_path("smk.config", "medit_guide_pred.yaml")
	cluster_template_path = project_file_path("smk.config", "medit_cluster.yaml")
	with open(config_template_path, 'r') as config_handle:
		config_template = yaml.safe_load(config_handle)
	with open(cluster_template_path, 'r') as cluster_handle:
		cluster_template = yaml.safe_load(cluster_handle)

	#   == Load Database configuration file generated by db_set ==
	try:
		with open(config_db_path, 'r') as config_db_handle:
			config_db = yaml.safe_load(config_db_handle)
	except FileNotFoundError:
		prRed(f"Couldn't fund the file {config_db_path}. "
		      f"Please double-check the path to <medit_database> and provide the path via '-d' option")

	# === Assign Variables to Configuration File ===
	config_template['run_name'] = f"{mode}_{jobtag}"
	config_template['support_tables'] = db_path_full
	config_template['processing_mode'] = mode
	config_template['output_directory'] = root_dir
	config_template['variant_query_dir'] = query_input
	# Assign run parameters to config
	config_template['qtype'] = qtype
	config_template['editor_request'] = editor_request
	config_template['be_request'] = be_request
	# Assign cluster options
	cluster_template['__default__']['cores'] = ncores
	cluster_template['__default__']['time'] = maxtime

	# ->=== PRIVATE GENOME RUN ===<-
	# == Check run mode ==
	if mode == 'private':
		# == Enforce presence of private genome in this mode ==
		if not private_genomes:
			print("Please provide a VCF input file to run mEdit's private mode")
			sys.exit(1)
		# == Create a private VCF directory when a private run is issued
		vcf_dir_path = f"{config_db['meditdb_path']}/{mode}/source_vcfs"
		set_export(vcf_dir_path)
		# == Check the config_db file for instructions on which human genome version should be used
		latest_reference = bool(config_db["latest_reference"])
		custom_reference = bool(config_db["custom_reference"])
		if latest_reference:
			config_template['sequence_id'] = ["latest_hg38"]
		if custom_reference:
			config_template['sequence_id'] = ["custom_reference"]
		# == VCF ID adjustment for private vcf run ==
		#   => Import VCF file prefix information to config file
		tagged_genomes = []
		count_tag = 1
		for private_genome in private_genomes:
			loop_tag = f"{jobtag}_{count_tag}"
			vcf_filename = f"{loop_tag}.vcf"
			tagged_genomes.append(loop_tag)
			#   => Avoid re-creating private VCF mirrors within the medit DB
			if not file_exists(f"{vcf_dir_path}/{vcf_filename}"):
				if not file_exists(f"{vcf_dir_path}/{vcf_filename}.gz"):
					#   => Create a copy of the VCF in the internal mEdit directory
					launch_shell_cmd(f"cp {private_genome} {vcf_dir_path}/{vcf_filename}", True)
					#   => Check VCF file compression and compress if necessary
					compress_file(f"{vcf_dir_path}/{vcf_filename}")
			count_tag += 1
		#   => Add any amount of private genomes to the config file
		config_template["vcf_id"] = tagged_genomes

	# === Write YAML configs to mEdit Root Directory ===
	write_yaml_to_file(config_template, dynamic_config_path)
	write_yaml_to_file(cluster_template, dynamic_cluster_path)

	# === Invoke SMK Pipelines ===

	print("# == Calling Guide Prediction pipeline == #")
	for smk_setup_idx in range(len(allowed_rules)):
		try:
			# --> When cluster submission is switched on,
			launch_shell_cmd(f"snakemake "
			                 f"--snakefile {project_file_path('smk.pipelines', 'guide_prediction.smk')} "
			                 # f"--directory {project_file_path('smk.pipelines', 'guide_prediction.smk')} "
			                 f"-j {ncores} "
			                 f"{smk_run_triggers} "
			                 f"{allowed_rules[smk_setup_idx]} "
			                 f"{cluster_smk_setup[smk_setup_idx]} "
			                 f"--configfile {config_db_path} "
			                 f"{dynamic_config_path} "
			                 f"--use-conda "
			                 f"{dryrun_setup}",
			                 smk_verbosity[smk_setup_idx]
			                 )
		except subprocess.CalledProcessError as e:
			print(f"Error: {e}")
		except ValueError:
			print(f"Process completed in a previous run. Moving to the next one...")



# == Native Modules

# == Installed Modules
from setuptools import setup, find_packages
# == Project Modules


setup(
    name='meditability',
    version='0.2.1',
    description='',
    author='Interventional Genomics Unit',
    author_email='',
    entry_points={
        "console_scripts": [
            "medit = prog:main"
        ]
    },
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"smk.config": ["*.yaml"],
                  "smk.envs": ["*.yaml"],
                  "smk.pipelines": ["*.smk"]},
    include_package_data=True,
    install_requires=[
        'snakemake>=7.32.4',
        'pulp<2.8.0',
        'wrapt>=1.15.0',
        'biopython>=1.81',
        'pyyaml>=6.0',
        'importlib-resources>=6.1.1',
        'pytz>=2023.3',
        'boto3>=1.28.57',
        'alive-progress>=3.1.5'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
    ],
)



# == Native Modules
import pickle
from os.path import abspath
# == Installed Modules
import yaml
# == Project Modules


def ls_editors(args):
	# == Load Run Parameters values ==
	print_editors = args.editors
	print_base_editors = args.base_editors

	# === Load Database Path ===
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	config_db_path = f"{db_path_full}/config_db/config_db.yaml"

	# === Load configuration file ===
	with open(config_db_path, 'r') as config_handle:
		config_db_obj = yaml.safe_load(config_handle)

	# === Load Editors Lists From Path Specified on config_db.yaml ===
	if print_editors:
		with open(str(config_db_obj["editors"]), 'rb') as editors_pkl_handle:
			editors_dict = pickle.load(editors_pkl_handle)
			print(f"Available editors:\n {list(editors_dict['clinical'].keys())}")
	if print_base_editors:
		with open(str(config_db_obj["base_editors"]), 'rb') as be_pkl_handle:
			base_editors_dict = pickle.load(be_pkl_handle)
			print(f"Available base editors:\n {list(base_editors_dict['all'].keys())}")



# Native Modules
import pickle
import subprocess
# Installed Modules
import pandas as pd
import vcf
# == Pysam REQUIRED ==
# Project Modules
from dataH import DataHandler

#################################
# searches for changes in guides found in hg38 based on variants in a user submitted VCF
#################################

## assumption that the VCF imported. Has the Alternate gene specified
## A deletion greater than 5


def extract_vcf_record(snv_coord, vcf_fname, window=30):
    '''
    seraches for an alternate genome variant in the +/-60bp of snv query position
    '''
    #snv_coord = 'chr11:5226788'
    #extracted_seq = 'ATCCCCAAAGGACTCAAAGAACCTCTGGGTTCAAGGGTAGACCACCAGCAGCCTAAGGGT'
    ch, pos = snv_coord.split(':')
    records_found = []

    vcf_reader = vcf.Reader(filename = vcf_fname, compressed = True)
    for record in vcf_reader.fetch(ch, int(pos) - window, int(pos) + window):
        records_found.append(record)
        print(record)
    return records_found


def extract_variant_info(record, hg38extracted_seq, hg38coord):
    '''
    for a given alt record found determine the ALT information
    and create a new extracted sequence incorporating the ALT allele
    '''
    vt = record.var_type
    x = record.samples[0]
    alt_alleles = record.ALT
    #print(record.ALT,record.REF)
    zyg = 'heterozygous' if x.is_het else 'homozygous'
    if len(alt_alleles) == 2:
        zyg = zyg + '-biallelic'
    if len(alt_alleles) > 2:
        zyg = zyg + '-multiallelic'

    alt = alt_alleles[0]
    hg38_start = int(hg38coord.split(':')[1]) - int(len(hg38extracted_seq)/2)
    vstart, vend = record.start, record.end #reference range
    rel_pos = ((vstart -hg38_start)+ 1,(vend-hg38_start)+1)
    #print(rel_pos)
    #print(vstart,vend)
    #print(alt.type)
    if len(record.REF) == len(alt): ## substitution
        vt = vt + '-sub'
        new_seq = hg38extracted_seq[0:rel_pos[0]] + alt.sequence + hg38extracted_seq[rel_pos[1]:]
        #print(new_seq)

    elif len(record.REF) > len(alt): # deletion
        vt = vt + '-del'
        #new_seq = hg38extracted_seq[0:rel_pos[0]] + alt.sequence + hg38extracted_seq[rel_pos[1]:]
        new_seq = 'undetermined'

    elif len(record.REF) < len(alt): # insertion
        vt = vt + '-ins'
        new_seq = hg38extracted_seq[0:rel_pos[0]] + alt.sequence + hg38extracted_seq[rel_pos[1]:]
        new_seq = new_seq[0:len(hg38extracted_seq)-1]
        #print(new_seq)

    else:
        new_seq = 'undetermined'

    return record.REF, alt.sequence, zyg, vt, new_seq


def find_overlapping_variants(vcf_fname, altgenome_name, models_path, hg38_snvinfo, search_params):

    new_guides = {}
    v_info = [[],[],[],[],[],[],[]]
    for ch, data in hg38_snvinfo.items():
        for d in data:
            try:
                query, tid, eid, strand, hgref, hgalt, feature_annotation, hg38extracted_seq, codons, hg38coord = d
            except ValueError:
                print(f"WARNING: The query below has the wrong number of values to unpack. "
                      f"Needs further investigation:\n{d}")
                continue

            # Check VCF if variant exsists in hg38 extracted_seq
            records_found = extract_vcf_record(snv_coord=hg38coord, vcf_fname = vcf_fname, window=30)

            if len(records_found) > 0:
                # Create new variant incorporated extracted seq
                for rec in records_found:
                    ref, alt, zyg, vtype, var_seq = extract_variant_info(rec,str(hg38extracted_seq),hg38coord)
                    v_info[0].append(query)
                    v_info[1].append(vtype)
                    alts = ",".join([str(x) for x in rec.samples[0].site.alleles][1:])
                    v_info[2].append(f'{ref}|{alts}')
                    v_info[3].append(alt)
                    v_info[4].append(f'chr{ch}:{rec.POS}')
                    v_info[5].append(zyg)
                    if var_seq != 'undetermined':
                        dh = DataHandler(query, strand, hgref, hgalt, feature_annotation, var_seq, codons, hg38coord)
                        new_guides_set, be_none = dh.get_Guides(search_params)
                        if len(new_guides_set['gRNA']) > 0:
                            if len(new_guides.keys()) == 0:
                                for k, v in new_guides_set.items():
                                    new_guides[k] = v
                            else:
                                for k, v in new_guides_set.items():
                                    new_guides[k] += v
                        v_info[6].append('placeholder')
                    else:
                        v_info[6].append('undetermined')

    labels = ['QueryTerm',f'{altgenome_name}_Variant_Type','REF|ALT','Examined_ALT','Var_Position', f'{altgenome_name}_Zygosity',f'{altgenome_name}_guide_impact']
    variants_found = dict(zip(labels,v_info))

    return variants_found, new_guides


def write_results(hg38guide_results, variants_found, new_guides_dict, altgenome_name, refgenome_name, outfile):
    '''
    compares guides found in new genome and ref genome.
    Drops unchanged guides and labels guides impacted by ALT
    '''

    if len(variants_found['QueryTerm']) > 1:
        hg38_gdf = pd.read_csv(hg38guide_results)
        var_df = pd.DataFrame(variants_found).set_index('QueryTerm')
        new_guides = pd.DataFrame(new_guides_dict).drop(columns=[ 'Guide_ID','SNV Position', 'Ref>Alt','Annotation'])
        hg38_gdf = hg38_gdf.loc[hg38_gdf.QueryTerm.isin(set(variants_found['QueryTerm']))]
        hg38_gdf = hg38_gdf.drop(columns=['SNV Position', 'Ref>Alt'])
        old_guides = hg38_gdf.join(var_df, how='outer', on='QueryTerm').reset_index(drop=True)

        new_info = new_guides.to_dict('tight')['data']
        print('            ')
        print('Guide Impacts')
        new_rows = []
        for idx, row in old_guides.iterrows():
            cnt = 1
            query, ed, coord, grna, pam = row['QueryTerm'],row['Editor'],row['Coordinates'],row['gRNA'],row['Pam']
            print('--------------------------------------')
            print('QUERY Editor QUERY_POS REF|ALT')
            print(query, ed, coord, row['REF|ALT'])
            if row[f'{altgenome_name}_guide_impact'] == 'undetermined':  # guide never determined (ALT is deletion)
                new_rows.append(list(row[0:8]) + ['-','-','-','undetermined'] + list(row[9:14]))
                print('impact;undetermined')
                cnt -= 1
            else:
                for n in new_info:
                    if ed == n[1]:
                        if [query,pam,grna] == [n[0],n[5],n[4]]:  # unchanged
                            print('impact;unchanged', pam, '->', n[5], grna, '->', n[4])
                            new_info.remove(n)
                            cnt -= 1
                            break
                        elif [query,pam] == [n[0],n[5]]:  # same pam which means change is in grna
                            new_rows.append(list(row[0:8]) + [n[4], n[5], n[6], 'grna_changed_conserved'] + list(row[9:14]))
                            print('impact;grna_changed_conserved', pam, '->', n[5], grna, '->', n[4])
                            new_info.remove(n)
                            cnt -= 1
                            break
                        elif[query,grna] == [n[0],n[4]]:
                            new_rows.append(
                                list(row[0:8]) + [n[4], n[5], n[6], 'pam_changed_conserved'] + list(row[9:14]))
                            print('impact;pam_changed_conserved', pam, '->', n[5], grna, '->', n[4])
                            new_info.remove(n)
                            cnt -= 1
                            break
                        else:
                            pass
                if cnt < 0:  # old guides remaining and not matched == no longer exsist
                    new_rows.append(list(row[0:8]) + ['removed', 'removed', 'removed', 'pam_changed_removed'] + list(row[9:14]))
                    print('impact;pam_changed_removed', pam, '->', '-', grna, '->', '-')

        if len(new_info) > 0:  # new guides remaining and not matched == new guides are made
            id_cnt = 1
            for n in new_info:
                x = [n[0],n[1],f'{n[1]}_NEW{id_cnt}',n[2],n[3],'-', '-', '-','-',
                                 n[4], n[5], n[6]] + list(var_df.loc[n[0]])
                new_rows.append(x[:-1])
                print('impact;pam_changed_added', '-', '->', n[5], '-', '->', n[4])
                id_cnt += 1
        cols = ['QueryTerm', 'Editor', 'Guide_ID', 'Coordinates', 'Strand', f'{refgenome_name}_gRNA',
       f'{refgenome_name}_Pam', f'{refgenome_name}_Doench Score',f'{refgenome_name}_gRNA',
       f'{altgenome_name}_Pam', f'{altgenome_name}_Doench Score',f'{altgenome_name}_Guide_Impact',
        'Variant_Type', 'REF|ALT','Examined_ALT', f'{altgenome_name}Var_Position', f'{altgenome_name}_Zygosity']
        df = pd.DataFrame(new_rows, columns=cols)
        df.to_csv(outfile, index=False)


def fetch_ALT_guides(vcf_fname, ref_guide_report, searchp_path, models_path, sitep_path, diffguides_report, altgenome_name, refgenome_name):

    # Get search parameters and the results from the reference assembly
    hg38guide_results = ref_guide_report
    search_params = pickle.load(open(searchp_path, 'rb'))
    # get hg38_snv_info (60bp extracted sequence, translation info, hg38_coordinates etc.
    hg38_snvinfo = pickle.load(open(sitep_path, 'rb'))

    variants_found, new_guides_dict = find_overlapping_variants(vcf_fname, altgenome_name, models_path, hg38_snvinfo, search_params)

    if len(variants_found['QueryTerm']) > 0:
        write_results(hg38guide_results,
                      variants_found,
                      new_guides_dict,
                      altgenome_name,
                      refgenome_name,
                      diffguides_report)
    else:
        print(f'no overlapping variants detected in {altgenome_name}')
        with open(diffguides_report, 'w') as f:
            f.write(f"No guide differences found based on the VCF {altgenome_name}")


def main():
    # SNAKEMAKE IMPORTS
    # === Inputs ===
    filtered_vcf = str(snakemake.input.filtered_vcf)
    guides_report = str(snakemake.input.guides_report_out)
    guide_search_params = str(snakemake.input.guide_search_params)
    snv_site_info = str(snakemake.input.snv_site_info)
    # === Outputs ===
    diffguides_out = str(snakemake.output.diff_guides)
    # === Params ===
    idx_filtered_vcf = str(snakemake.params.idx_filtered_vcf)
    # ==* The models_path ideally should not be here.
    #       Once the DataHandler class is properly instantiated this can be removed
    models_path = str(snakemake.params.models_path)
    # === Wildcards ===
    altgenome_name = str(snakemake.wildcards.vcf_id)
    refgenome_name = str(snakemake.wildcards.sequence_id)

    # resultsfolder = "/groups/clinical/projects/editability/medit_queries/medit_test/test_out/"
    # vcf_fname = "/groups/clinical/projects/editability/tables/raw_tables/VCFs/HG02257.filtered.vcf.gz"
    # altgenome_name = 'HG02257'
    # refgenome_name = 'HG38'

    # Generate vcf index with tabix
    print(f"Generate tabix file on:\n {idx_filtered_vcf}")
    subprocess.run(f"tabix {filtered_vcf}", shell=True)

    fetch_ALT_guides(filtered_vcf,
                     guides_report,
                     guide_search_params,
                     models_path,
                     snv_site_info,
                     diffguides_out,
                     altgenome_name,
                     refgenome_name)


if __name__ == "__main__":
    main()



# Installed Modules
from Bio.Seq import Seq
from Bio import SeqUtils
from Bio.SeqUtils import seq3
# Project Modules
import scoring


class DataHandler:
    """
    search for guides given a genomic sequence and SNV info
    """

    def __init__(self, query, strand, ref, alt, feature_annotation, models_dir, extracted_seq, rf, coord,gname):
        """
        :param query: ex: 'NM_000532.5(PCCB):c.1316A>G (p.Tyr439Cys)' or 'chr19:136327650A>G'
        :param strand: ex. '-' or '+
        :param ref: ex. 'A'
        :param alt: ex. 'G'
        :param feature_annotation: ex: 'exon'
        :param extracted_seq: 'CCCACAGGGCCCTCACCTGCAGATTGTGATTGTGGCCGCACAGGTAGGCAGTGACCCCGT'
        :param rf : ex: '2'
        :param coord : ex: 'chr19:136327650'
        """
        # search data
        self.NC_ref_allele = str(ref).upper()
        self.NC_alt_allele = str(alt).upper()
        self.strand = strand  # coding_strand
        self.NM_ref_allele = self.NC_ref_allele if self.strand == '+' else str(Seq(self.NC_ref_allele).complement())
        self.NM_alt_allele = self.NC_alt_allele if self.strand == '+' else str(Seq(self.NC_alt_allele).complement())
        self.SNV_chr_pos = int(coord.split(':')[1])
        self.query = query
        self.rf = rf
        self.extracted_seq = str(extracted_seq)
        self.annotation = feature_annotation
        self.models_dir = models_dir
        self.coord = coord
        self.chrom = coord.split(':')[0].replace('chr', '')
        self.gname = gname

        # search params
        self.pam = str()  # Ex. 'NGG'
        self.pamISfirst = False  # Boolean
        self.win_size = list()  # Ex. list [4,8]
        self.gscoring = None  # Ex. True/False
        self.guidelen = 20

        # outputs
        self.guides_found = {'QueryTerm': [], 'GeneName':[],'Editor': [], 'Guide_ID': [], 'Coordinates': [],
                             'Strand': [], 'gRNA': [], 'Pam': [], 'SNV Position': [],
                             'On-Target Efficiency Score': [], 'OOF Score':[],'Ref>Alt': [], 'Annotation': []}

        self.BEguides_found = {'QueryTerm': [], 'GeneName':[],'Base Editor': [], 'Guide_ID': [], 'Coordinates': [],
                               'Strand': [],'gRNA': [], 'Pam': [], 'SNV Position': [],
                               'Hg38 Reference (Codon>AA)': [], 'Alternate (Codon>AA)': [], 'BE Converted (Codon>AA)': [],
                               'Conversion Type': [], 'Bystander': [], 'Annotation': []}


    def set_guide_search_params(self, pam, pamISfirst, win_size, gscoring, guidelen):
        self.pam = pam
        self.pamISfirst = pamISfirst
        self.win_size = win_size
        self.gscoring = gscoring
        self.guidelen = guidelen

    def find_codon(self, snv_rel_pos):
        if self.strand == '+':
            codon = self.extracted_seq[int(snv_rel_pos - self.rf): int((snv_rel_pos - self.rf) + 3)]
        else:
            adj_rf = 2 - self.rf
            codon = self.extracted_seq[int(snv_rel_pos - adj_rf): int(snv_rel_pos - adj_rf)+3]
            codon = Seq(codon).reverse_complement()
        return codon

    @staticmethod
    def get_AAconversion_type(codon1, codon2):
        """
        codon1: codon of Alt allele to be changed by BE
        codon2: codon after changed by BE
        """
        aa_groups = [["G", "A", "V", "L", "I", "M", "F", "Y", "W"],
                     ["S", "Q", "T", "N"],
                     ["C", "G", "P"],
                     ["D", "E"],
                     ["K", "H", "R", "Q"]]
        codon1, codon2 = Seq(codon1), Seq(codon2)
        aa1, aa2 = codon1.translate(), codon2.translate()
        mtype = ""
        if aa1 == aa2:
            mtype = 'Synonymous'
            if codon1 == codon2:
                mtype = 'Silent'
        else:
            if codon2 in ['TAA', 'TAG', 'TGA']:
                mtype = 'Nonsense'
            elif ([aa_groups.index(x) for x in aa_groups if str(aa1) in x] ==
                  [aa_groups.index(x) for x in aa_groups if str(aa2) in x]):
                mtype = 'Conservative'
            else:
                mtype = 'Non-conservative'
        return mtype

    def getBE(self, guides, conversion, win_size, name):
        """
        Finds codon level SNV and determines if the Base Editor Conversion can work
        """
        coding_strand = self.strand
        snv_rel_pos = len(self.extracted_seq) / 2

        for i in range(len(guides)):
            editor, guide, pam_found, guide_strand, snvpos, on_score, oof_score, start, end = guides[i]

            target_bases = Seq(guide[win_size[0] - 1:win_size[1] + 1])  # Bases inside 4-8 window

            # Converted case
            convert = str(conversion[1])
            bystander = target_bases.count(conversion[0]) - 1
            if self.annotation not in ['exon','start_codon','stop_codon']:
                ctype = 'NA'
                self.add_BEguides(name,
                                  guide,
                                  pam_found,
                                  guide_strand,
                                  snvpos,
                                  start,
                                  end,
                                  self.NM_ref_allele,
                                  self.NM_alt_allele,
                                  convert,
                                  ctype,
                                  bystander)

            else:
                ## In Exon
                #Determine codon and translated product
                alt_codon = self.find_codon(snv_rel_pos)

                # Alternative codon
                aa_alt = Seq(alt_codon).translate()

                # Reference codon
                ref_codon = Seq("".join(alt_codon[x] if x != abs(self.rf) else self.NM_ref_allele for x in [0,1,2]))
                aa_ref = ref_codon.translate()

                ##Converted allele
                convert = convert if self.NM_alt_allele == conversion[0] else str(Seq(convert).complement())

                new_codon = Seq("".join(alt_codon[x] if x != abs(self.rf) else convert for x in [0,1,2]))
                aa_new = new_codon.translate()

                mtype = self.get_AAconversion_type(codon1=ref_codon, codon2=new_codon)

                ### If conversion leads to Ref change or REf change keep
                ctype = mtype
                aa_new = seq3(aa_new, custom_map={"*": "***"})
                aa_alt = seq3(aa_alt, custom_map={"*": "***"})
                aa_ref = seq3(aa_ref, custom_map={"*": "***"})
                ref = f"{ref_codon}>{aa_ref}"
                alt = f"{alt_codon}>{aa_alt}"
                convert = f"{new_codon}>{aa_new}"
                self.add_BEguides(name,
                                  guide,
                                  pam_found,
                                  guide_strand,
                                  snvpos,
                                  start,
                                  end,
                                  ref,
                                  alt,
                                  convert,
                                  ctype,
                                  bystander)

    def add_guides(self, name, guide, pam_found, strand, snvpos, on_score,oof_score,start,end):
        self.guides_found['QueryTerm'].append(self.query)
        self.guides_found['GeneName'].append(self.gname)
        self.guides_found['Editor'].append(name)
        self.guides_found['Guide_ID'].append(f'{name}_')
        self.guides_found['Coordinates'].append(f'{self.chrom}:{start}-{end}')
        self.guides_found['On-Target Efficiency Score'].append(on_score)
        self.guides_found['OOF Score'].append(oof_score)
        self.guides_found['Strand'].append(strand)
        self.guides_found['Pam'].append(str(pam_found))
        self.guides_found['gRNA'].append(str(guide))
        self.guides_found['Ref>Alt'].append(f"{self.NM_ref_allele}>{self.NM_alt_allele}")
        self.guides_found['SNV Position'].append(snvpos)
        self.guides_found['Annotation'].append(self.annotation)

    def add_BEguides(self, name, guide, pam_found, strand, snvpos, start, end, ref, alt, convert, ctype, bystander):
        self.BEguides_found['QueryTerm'].append(self.query)
        self.BEguides_found['GeneName'].append(self.gname)
        self.BEguides_found['Guide_ID'].append(f'{name}_')
        self.BEguides_found['Base Editor'].append(name)
        self.BEguides_found['Coordinates'].append(f'{self.chrom}:{start}-{end}')
        self.BEguides_found['gRNA'].append(str(guide))
        self.BEguides_found['Pam'].append(str(pam_found))
        self.BEguides_found['SNV Position'].append(snvpos)
        self.BEguides_found['Strand'].append(strand)
        self.BEguides_found['Hg38 Reference (Codon>AA)'].append(ref)
        self.BEguides_found['Alternate (Codon>AA)'].append(alt)
        self.BEguides_found['BE Converted (Codon>AA)'].append(convert)
        self.BEguides_found['Conversion Type'].append(ctype)
        self.BEguides_found['Bystander'].append(bystander)
        self.BEguides_found['Annotation'].append(self.annotation)

    def get_guide_set(self, name, pam, pamISfirst, win_size, guidelen, BEmode):
        """
        :param name:
        :param pam: pam seq ex:'NGG'
        :param pamISfirst: 5'or3'PAM ex:True/False
        :param win_size: list containing upper and lower limits of the targetable window. Ex:[4,8]
        :param search window: intial search + or - SNV site
        :param guidelen: guide without pam length
        :return: Guide Dictionary
        """
        guides = []
        pamlen = len(pam)
        sitelen = guidelen + pamlen
        snv_rel_pos = int(len(self.extracted_seq)/2)
        on_score, oof_score = '-','-'
        if BEmode:
            win_size = [win_size[0] - guidelen -1,win_size[1] -guidelen-1]

        pam_min, pam_max = int((snv_rel_pos - win_size[1]))- 1, int((snv_rel_pos - win_size[0])) - 1

        if pamISfirst == True:
            pam_min, pam_max = pam_min + pamlen, pam_max + pamlen

        # Narrow based on guide params
        for search_strand in ["-", "+"]:
            search_seq = Seq(self.extracted_seq) if search_strand == "+" else Seq(self.extracted_seq).reverse_complement()
            pam_index = SeqUtils.nt_search(str(search_seq), pam)[1:]

            for i in pam_index:
                if i in range(pam_min, pam_max + 1):

                    if not pamISfirst:  # 3' PAM
                        target_start = i - guidelen
                        guide = search_seq[i - guidelen:i]
                        if not BEmode:
                            if pam == 'NGG' and guidelen == 20:
                                #Azmith only accurate for NGG pams
                                on_score = scoring.azimuth([str(search_seq[target_start - 3:target_start + sitelen + 4])],
                                                           self.models_dir)[0]
                            #oof_score use only for DSB
                            mh_score, oof_score = scoring.oofscore(str(search_seq[target_start - 20:target_start + sitelen + 20]))

                        pam_found = str(search_seq[i:i + pamlen])

                    else:
                        target_start = i + pamlen
                        guide = search_seq[target_start: i + sitelen]
                        pam_found = search_seq[i:target_start]
                        if 'Cas12a' in name:
                            on_score = round(scoring.deepcpf1([str(search_seq[i - 5:i + sitelen + 4])],
                                                              self.models_dir)[0][0], 2)

                    snvpos = snv_rel_pos - target_start
                    start = self.SNV_chr_pos - snvpos
                    end = start + sitelen

                    guides.append([name, guide, pam_found, search_strand, snvpos, on_score, oof_score, start, end])

                    if not BEmode:
                        self.add_guides(name, guide, pam_found, search_strand, snvpos, on_score, oof_score, start, end)
        return guides

    def get_Guides(self, search_params, BEsearch_params=None):

        for name, params, in search_params.items():
            pam, pamISfirst, guidelen, dsb_loc = params[0:4]
            win_size = [int(dsb_loc)-7, int(dsb_loc)+7]
            guides = self.get_guide_set(name, pam, pamISfirst,win_size, guidelen, BEmode=False)

        # if BE mode is on
        if BEsearch_params is not None:
            for k, params, in BEsearch_params.items():
                pam, pamISfirst, guidelen, win_size = params[0][0], params[0][1], params[0][2], params[0][3]
                bguides = self.get_guide_set(k, pam, pamISfirst, win_size, guidelen, BEmode=True)

                # if guides are found sep neg and pos strand guides
                if len(bguides) > 0:
                    pos_guides, neg_guides = [], []
                    for g in bguides:
                        if g[3] == '+':
                            pos_guides += [g]
                        else:
                            neg_guides += [g]

                    # See if SNV can be BE edited
                    for p in range(1, len(params[1:]) + 1):
                        conversion = params[p][0]  # 'CT'
                        name = ",".join(
                            [n for n in params[p][1:]])

                        if self.NC_alt_allele == conversion[0]:
                            if len(pos_guides) > 0:
                                self.getBE(guides=pos_guides, conversion=conversion, win_size=win_size, name=name)

                        if self.NC_alt_allele == str(Seq(conversion[0]).complement()):
                            if len(neg_guides) > 0:
                                self.getBE(guides=neg_guides, conversion=conversion, win_size=win_size, name=name)

        return self.guides_found, self.BEguides_found


'''
#----------------------------Test----------------------
datadir = "/groups/clinical/projects/editability/tables/"
processed_tables = "/groups/clinical/projects/editability/tables/processed_tables/"
fasta_path ="/groups/clinical/projects/clinical_shared_data/hg38/hg38.fa.gz"

search_params= {'spCas9': ('NGG', False, 20, -2, 'Sp Cas9, SpCas9-HF1, eSpCas9 1.1'),
                'saCas9': ('NNGRRT', False,21, -2, 'Cas9 S. Aureus 21 base guide'),
                'CasX': ('TTCN', True, 20, 18, 'Cas12e'),
                'AsCas12a': ('TTTV', True, 23, 22, 'TTT(A/C/G)-23bp - Cas12a (Cpf1)')}
BE_search_params = {'spCas9-def': [('NGG', False, 20, [4, 8]), ('CT', 'BE3', 'BE4', 'BE4max', 'BE4-Gam'), ('AG', 'ABE7.9', 'ABE7.10', 'ABEmax')]}

snv_info = {'11': [['NM_000518.5:c.114G>A', '-', 'C', 'T', 'exon', Seq('ATCCCCAAAGGACTCAAAGAACCTCTGGGTTCAAGGGTAGACCACCAGCAGCCTAAGGGT'), -2, 'chr11:5226778']], 
            '3': [['NM_000532.5:c.1316A>G', '+', 'A', 'G', 'exon', Seq('TGGATCTGTTTTAGGCCTATGGAGGTGCCTGTGATGTCATGAGCTCTAAGCACCTTTGTG'), 1, 'chr3:136327650']],
            '16': [['NM_000517.6:c.99G>A', '+', 'G', 'A', 'exon', Seq('CACCCCTCACTCTGCTTCTCCCCGCAGGATATTCCTGTCCTTCCCCACCACCAAGACCTA'), 2, 'chr16:173128'], 
                   ['NM_005886.3:c.1A>G', '+', 'A', 'G', 'exon', Seq('GTGGGGCTTCAGGTGCCAGCCAGCTGAAGGGTGGCCACCCCTGTGGTCACCAAGACAGCC'), 2, 'chr16:57737244']]}

for ch, data in fg.snv_info.items():
    for d in data:
        query, tid, eid, strand, ref, alt, feature_annotation, extracted_seq, codons, coord = d
        print(f'----------{query}--------------')
        dh = DataHandler(query, strand, ref, alt, feature_annotation, extracted_seq, codons, coord)
    guides_found, BEguides_found = dh.get_Guides(fg.search_params,fg.BE_search_params)
        print(len(guides_found['gRNA'])
            #for k,v in guides_found:
    #    print(k,v)
    #for k,v in BEguides_found.items():
    #    print(k,v)

'''
